#include <stdio.h>
// #include <X11/Xlib.h>
// #include <X11/Xutil.h>
// #include <X11/keysym.h>

//#define __hpux

#define NAME "A Simple Window"
#define ICON_NAME "Simple"
char *STRING= "Hello World";
char *PRESSED= "MOUSE BUTTON pressed";
#ifdef __hpux
#define FONT "vbee-36"
#else
#define FONT "r14"
#endif


class G__XWindow {
 public:
  enum XSTATUS { NOWINDOW, EXPOSED };
  G__XWindow(int xsize=500,int ysize=500) { 
    xstatus = NOWINDOW;
    createWindow(xsize,ysize); 
  }
  ~G__XWindow() {  closeWindow(); }
  int xrawsize() { return(xwa.width); }
  int yrawsize() { return(xwa.height); }
  int xstatus;

 protected:
  XWMHints xwmh ;
  unsigned int fontheight,pad,fg,bg,bd,bw;
  Display *dpy;
  Window win;
  GC gc;
  XFontStruct *fontstruct;
  XEvent event;
  XSizeHints xsh;
  XWindowAttributes xwa;
  XSetWindowAttributes xswa;

  void setXWMHints();
  void createWindow(int xsize,int ysize);
  void closeWindow();
};


void G__XWindow::closeWindow()
{
  XUnmapWindow(dpy,win);
  XDestroyWindow(dpy,win);
  XCloseDisplay(dpy);
  printf("\n");
}


void G__XWindow::setXWMHints()
{
  xwmh.flags = InputHint|StateHint;
  xwmh.input = False;
  xwmh.initial_state = NormalState;
  xwmh.icon_pixmap=0;
  xwmh.icon_window=0;
  xwmh.icon_x=0;
  xwmh.icon_y=0;
  xwmh.icon_mask=0;
  xwmh.window_group=0;
}

void G__XWindow::createWindow(int xsize,int ysize)
{

  int argc=0;
  char **argv;
  
  setXWMHints();
  
  if((dpy=XOpenDisplay(NULL))==NULL) {
    fprintf(stderr,"%s: can't open %s.\n"
	    ,"plot3d",XDisplayName(NULL));
    exit(1);
  }
  
  if((fontstruct = XLoadQueryFont(dpy,FONT))==NULL) {
    fprintf(stderr,"%s: display %s doesn't know font %s\n"
	    ,"plot3d",DisplayString(dpy),FONT);
    exit(1);
  }
  
  fontheight = fontstruct->max_bounds.ascent+fontstruct->max_bounds.descent;
  
  bd = WhitePixel(dpy,DefaultScreen(dpy));
  bg = BlackPixel(dpy,DefaultScreen(dpy));
  fg = WhitePixel(dpy,DefaultScreen(dpy));
  
  bw=1;
  pad = 1;
  
  xsh.flags = (PPosition|PSize);
  xsh.height = xsize ; // fontheight + 2 * pad;
  xsh.width = ysize; // XTextWidth(fontstruct,STRING,strlen(STRING))+2*pad;
  xsh.x = (DisplayWidth(dpy,DefaultScreen(dpy))-xsh.width)/2;
  xsh.y = (DisplayHeight(dpy,DefaultScreen(dpy))-xsh.height)/2;
  
  win = XCreateSimpleWindow(dpy,DefaultRootWindow(dpy),
			    xsh.x,xsh.y,xsh.width,xsh.height,bw,bd,bg);
  
  XSetStandardProperties(dpy,win,NAME,ICON_NAME,None,argv,argc,&xsh);
  XSetWMHints(dpy,win,&xwmh);
  
  xswa.colormap = DefaultColormap(dpy,DefaultScreen(dpy));
  xswa.bit_gravity = CenterGravity;
  XChangeWindowAttributes(dpy,win,(CWColormap|CWBitGravity),&xswa);
  
  gc = DefaultGC(dpy,DefaultScreen(dpy));
  XSetFont(dpy,gc,fontstruct->fid);
  XSetForeground(dpy,gc,fg);
  XSetBackground(dpy,gc,bg);
  
  XSelectInput(dpy,win 
	       ,ExposureMask | StructureNotifyMask |
	        KeyPressMask | ButtonPressMask |
	        KeyReleaseMask | ButtonReleaseMask |
	        PointerMotionMask
	       );
  
  XMapWindow(dpy,win);

}


/*
 * Local Variables:
 * c-tab-always-indent:nil
 * c-indent-level:2
 * c-continued-statement-offset:2
 * c-brace-offset:-2
 * c-brace-imaginary-offset:0
 * c-argdecl-indent:0
 * c-label-offset:-2
 * compile-command:"make -k"
 * End:
 */


#include <array.h>

/*************************************************************************
* 3 dimentional plot class
*************************************************************************/

class G__plot3d : public G__XWindow, public graphbuf {
// class G__plot3d : public graphbuf , public G__XWindow {

 public:
  void interaction();
  G__plot3d& operator <<(char c);
  G__plot3d& operator >>(char c);

 private:
  void drawgraph();
  double xmin,xmax,ymin,ymax;
  double xgradi,xgradn,ygradi,ygradn;

};

G__plot3d& G__plot3d::operator <<(char c)
{
  setStatus(G__GRAPHBUF_FIXED);
  interaction();
}

void G__plot3d::drawgraph()
{
  int i,j;
  int nplot ;
  int npoint; 
  double xmid,ymid;

  xmid = xrawsize()/2;
  ymid = yrawsize()/2;

  nplot = Nplot();
  for(j=1;j<nplot;j++) {
    npoint = Npoint(j);
    for(i=1;i<npoint;i++) {
      XDrawLine(dpy,win,gc
		,pdat[0][i-1],ymid-pdat[j][i-1]
		,pdat[0][i],ymid-pdat[j][i]);
    }
  }
}

void G__plot3d::interaction() 
{

  int xo,yo,xs,ys;
  int x,y;
  int prevx,prevy;
  int state=0;
  int k=0;

  if(EXPOSED==xstatus) drawgraph();
  xstatus = EXPOSED;
  
  while(1) {
    XNextEvent(dpy,&event);

    // Motion event
    if(event.type==MotionNotify) {
      if(state) {
	// dynamicmove(event.xmotion.x,event.xmotion.y);
	XDrawLine(dpy,win,gc,prevx,prevy,event.xmotion.x,event.xmotion.y);
	prevx=event.xmotion.x;
	prevy=event.xmotion.y;
      }
    }

    // Button events
    if(event.type==ButtonRelease) {
      state=0;
      if(XGetWindowAttributes(dpy,win,&xwa) ==0) break;
#ifdef BOX
      XClearWindow(dpy,win);
      xo = (x<event.xmotion.x)?x:event.xmotion.x;
      yo = (y<event.xmotion.y)?y:event.xmotion.y;
      xs = (int)fabs(x-event.xmotion.x);
      ys = (int)fabs(y-event.xmotion.y);
      XDrawRectangle(dpy,win,gc,xo,yo,xs,ys);
      x = (xwa.width-XTextWidth(fontstruct,STRING,strlen(STRING)))/2;
      y = (xwa.height+fontstruct->max_bounds.ascent-
	   fontstruct->max_bounds.descent)/2;

      XDrawString(dpy,win,gc,x,y,STRING,strlen(STRING));
      drawgraph();

      XCheckTypedEvent(dpy,Expose,&event);
      while(XCheckTypedEvent(dpy,Expose,&event));
#endif
    }

    if(event.type==ButtonPress) {
      if(XGetWindowAttributes(dpy,win,&xwa) ==0) break;
      x = event.xbutton.x;
      y = event.xbutton.y;
      prevx=x;
      prevy=y;
      state=1;
    }

    // Key events
    if(event.type==KeyRelease) {
      XClearWindow(dpy,win);
      return;
    }

    if(event.type==KeyPress) {
      // fprintf(stderr,"No%d\t%d\n",k,event.xkey.keycode);
      fprintf(stdout,"No%d\t%d\n",k++,event.xkey.keycode);
      if(event.xkey.keycode==15) break;
      XClearWindow(dpy,win);
    }

    // Size change and expose event
    if((event.type==ConfigureNotify) || (event.type==Expose) ) {
      if(XGetWindowAttributes(dpy,win,&xwa) ==0) break;
      x = (xwa.width-XTextWidth(fontstruct,STRING,strlen(STRING)))/2;
      y = (xwa.height+fontstruct->max_bounds.ascent-
	   fontstruct->max_bounds.descent)/2;
      XClearWindow(dpy,win);
      XDrawString(dpy,win,gc,x,y,STRING,strlen(STRING));
      drawgraph();
      while(XCheckTypedEvent(dpy,Expose,&event));
    }

  }

}

/*
 * Local Variables:
 * c-tab-always-indent:nil
 * c-indent-level:2
 * c-continued-statement-offset:2
 * c-brace-offset:-2
 * c-brace-imaginary-offset:0
 * c-argdecl-indent:0
 * c-label-offset:-2
 * compile-command:"make -k"
 * End:
 */

