/* /% C %/ */
/***********************************************************************
 * cint (C/C++ interpreter)
 ************************************************************************
 * Source file pause.c
 ************************************************************************
 * Description:
 *  Interactive interface
 ************************************************************************
 * Copyright(c) 1991~1996  [hp] Hewlett Packard Japan, All rights reserved
 * Author                  Masaharu Goto (gotom@jpn.hp.com)
 * Copyright(c) 1995~1997  Masaharu Goto (MXJ02154@niftyserve.or.jp)
 *
 * Permission to use, copy, modify and distribute this software and its 
 * documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.  Hewlett-Packard Japan Company makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 ************************************************************************/

#include "common.h"

#define STDIN_FILENO 0
#define STDOUT_FILENO 1
#define STDERR_FILENO 2

#define G__NUM_STDIN  0
#define G__NUM_STDOUT 1
#define G__NUM_STDERR 2

/**************************************************************************
* G__SET_TEMPENV
*
**************************************************************************/
#define G__SET_TEMPENV                                 \
      store_local = G__p_local;                        \
      store_struct_offset = G__store_struct_offset;    \
      store_tagnum=G__tagnum;                          \
      store_exec_memberfunc = G__exec_memberfunc;      \
      G__p_local=view_local;                           \
      G__store_struct_offset=view_struct_offset;       \
      G__tagnum=view_tagnum;                           \
      G__exec_memberfunc=view_exec_memberfunc

/**************************************************************************
* G__RESET_TEMPENV
*
**************************************************************************/
#define G__RESET_TEMPENV                               \
      G__p_local=store_local;                          \
      G__store_struct_offset=store_struct_offset;      \
      G__tagnum=store_tagnum;                          \
      G__exec_memberfunc=store_exec_memberfunc

#ifndef G__OLDIMPLEMENTATION696
/**************************************************************************
* G__STORE_EVALENV
*
**************************************************************************/
#define G__STORE_EVALENV                                  \
      struct G__input_file eval_ifile;                    \
      struct G__input_file eval_view;                     \
      struct G__var_array *eval_local=G__p_local;         \
      long eval_struct_offset=G__store_struct_offset;     \
      int eval_tagnum=G__tagnum;                          \
      int eval_exec_memberfunc=G__exec_memberfunc;        \
      int store_step=G__step;                             \
      eval_ifile=G__ifile;                                \
      eval_view=view;                                     \
      if(strncmp("s",com,1)==0) G__step=1;                \
      else if(strncmp("S",com,1)==0) G__step=0;           \
      else if(G__stepover) G__step=0

/**************************************************************************
* G__RESTORE_EVALENV
*
**************************************************************************/
#define G__RESTORE_EVALENV                                \
      view=eval_view;                                     \
      G__ifile=eval_ifile;                                \
      G__p_local=eval_local;                              \
      G__store_struct_offset=eval_struct_offset;          \
      G__tagnum=eval_tagnum;                              \
      G__exec_memberfunc=eval_exec_memberfunc;            \
      G__step=store_step

#endif



#ifdef G__ASM_DBG
extern int G__asm_step;
#endif

static int G__pause_return=0;
#ifdef G__OLDIMPLEMENTATION463
static FILE *fout=(FILE*)NULL;
#endif

#ifdef G__ROOT
static int G__rootmode=1;
#else
static int G__rootmode=0;
#endif

#ifndef G__OLDIMPLEMENTATION577
/************************************************************************
* G__display_keyword()
*
*  display keyword for '/[keyword]' debugger command
*
************************************************************************/
static void G__display_keyword(fout,keyword,fname)
FILE *fout;
char *keyword;
char *fname;
{
  char line[G__LONGLINE];
  char *null_fgets;
  FILE *keyfile;

  keyfile = fopen(fname,"r");

  if(keyfile) {
    null_fgets=fgets(line,G__LONGLINE-1,keyfile);
    while((char*)NULL!=null_fgets) {
      if(strstr(line,keyword)) {
	fprintf(fout,"%s",line);
	G__more_pause(fout);
      }
      null_fgets=fgets(line,G__LONGLINE-1,keyfile);
    }
    fclose(keyfile);
  }
  else {
    G__genericerror("Warning: can't open file. keyword search unsuccessful");
  }
}
#endif

#ifdef G__SECURITY
/******************************************************************
* G__security_recover()
******************************************************************/
void G__security_recover(fout)
FILE *fout;
{
  if(G__return>G__RETURN_NORMAL && G__security_error) {
    if((G__security_error&(G__DANGEROUS|G__FATAL))) {
      if(fout) {
#ifdef G__ROOT
	fprintf(fout,"*** Fatal error in interpreter... restarting interpreter ***\n");
#else
	fprintf(fout,"!!!Sorry, terminate cint session!!!\n");
#endif
      }
    }
    else {
      if(fout) {
#ifdef G__ROOT
	fprintf(fout,"*** Interpreter error recovered ***\n");
#else
	fprintf(fout,"!!!Error recovered!!!\n");
#endif
      }
      G__return=G__RETURN_NON;
      /* G__security_error = G__NOERROR; */ /* don't remember why not doing */
    }
  }
}
#endif

/******************************************************************
* Inevitable static variable
******************************************************************/
static struct G__input_file view;
static struct G__var_array *view_local,*store_local;
static long view_struct_offset,store_struct_offset;
static int view_tagnum,store_tagnum;
static int view_exec_memberfunc,store_exec_memberfunc;
#ifndef G__OLDIMPLEMENTATION495
static char stdoutsav[64];
static char stderrsav[64];
static char stdinsav[64];
#endif

/******************************************************************
* int G__init_process_cmd()
******************************************************************/
int G__init_process_cmd()
{
  view = G__ifile;
  view_local = G__p_local;
  view_struct_offset=G__store_struct_offset;
  view_tagnum=G__tagnum;
  view_exec_memberfunc=G__exec_memberfunc;
}

/******************************************************************
* int G__pause()
******************************************************************/
int G__pause()
{
#ifndef G__OLDIMPLEMENTATION432
  char *p;
  char cintname[G__ONELINE];
  char filename[G__ONELINE];
#endif
  char command[G__ONELINE];
  char prompt[G__ONELINE];
  int ignore=G__PAUSE_NORMAL;
  int more = 0;

  void (*oldhandler)();


#ifdef G__NEVER
  static int flag=0;
#endif

#ifndef G__OLDIMPLEMENTATION432
  p = strrchr(G__nam,G__psep[0]);
  if(p && *(p+1)) strcpy(cintname,p+1);
  else            strcpy(cintname,G__nam);
#endif

  prompt[0] = '\0';

  /************************************************
  * Do not pause while no execution
  ************************************************/
  if(G__no_exec!=0) return(ignore);

  /************************************************
  * Do not pause if breakfile is specified but
  * doesn't match to current source file
  ************************************************/
#ifdef G__OLDIMPLEMENTATION473
  if((G__step!=0)&&(G__break==0)&&(strcmp(G__breakfile,"")!=0)) {
    if(strcmp(G__breakfile,G__ifile.name)!=0) return(ignore);
  }
#endif

  /************************************************
  * Do not pause if assertion is set and the
  * condition is faulse
  ************************************************/
  if((G__step==0)&&(strcmp(G__assertion,"")!=0)&&
     (!G__test(G__assertion))) {
    return(ignore);
  }

  /************************************************
  * Prompt/comments printed out to stdout
  * Put caridge return
  ************************************************/
#ifdef G__OLDIMPLEMENTATION463
  fout=G__sout;
#endif
  fprintf(G__sout,"\n");


  G__init_process_cmd();

  /************************************************
  * If Display mode, display source code
  ************************************************/
  if((G__breakdisp!=0)&&(G__ifile.name[0]!='\0')) {
    G__pr(G__sout,view);
  }

  /************************************************
  * Infinite loop, until space or 'S'(Step)
  * command
  ************************************************/
  while(1) {

    /************************************
    * Set prompt string
    ************************************/
    if (prompt[0]) strcpy(command, prompt);
    else {
      if('\0'==view.name[0]) {
#ifndef G__OLDIMPLEMENTATION432
	sprintf(command,"%s> ", cintname);
#else
	sprintf(command,"%s> ", G__nam);
#endif
      }
      else {
#ifndef G__OLDIMPLEMENTATION432
	p = strrchr(view.name,G__psep[0]);
	if(p && *(p+1)) strcpy(filename,p+1);
	else            strcpy(filename,view.name);
	sprintf(command,"FILE:%s LINE:%d %s> "
		,filename,view.line_number,cintname);
#else
	sprintf(command,"FILE:%s LINE:%d %s> "
		,view.name,view.line_number,G__nam);
#endif
      }
    }

    if (more == 0) {
#ifdef G__ASM
       G__asm_noverflow=0;
#endif

    /************************************
    * flush files
    ************************************/
       G__disp_mask=0;
       fflush(G__sout);
       fflush(G__serr);
#ifdef G__DUMPFILE
       if(G__dumpfile!=NULL) fflush(G__dumpfile);
#endif

       /* if G__atpause is defined , evaluate the function */
       if(G__atpause) {
         (*G__atpause)();
       }

    /************************************
    * Exit program if Break key
    ************************************/
#ifndef G__ROOT /* This is only defined with ROOT */
       signal(SIGINT,G__killproc);
#endif

    /***********************************************
    * For ATEC replacement
    * If execute readline dump file mode, get
    * debug command from the readline dump file
    * else get command from stdin.
    ***********************************************/

       G__SET_TEMPENV;

       G__in_pause=1;

    } else {

#ifndef G__ROOT /* This is only defined with ROOT */
       signal(SIGINT,G__exit);
#endif

    }

    strcpy(command,G__input(command));

    if (more == 0) {
       G__in_pause=0;

       G__RESET_TEMPENV;
    }

    /************************************
    * G__pause() when break key
    ************************************/
    oldhandler = signal(SIGINT,G__breakkey);
    alarm(0);

    G__pause_return=0;

    if (ignore = G__process_cmd(command, prompt, &more)) break;
    if(G__pause_return) break;
  }

  G__pause_return=0;
#ifdef G__ROOT
  signal(SIGINT,oldhandler);
#endif
  return(ignore);
}

#ifndef G__OLDIMPLEMENTATION464
/******************************************************************
* G__update_stdio()
******************************************************************/
int G__update_stdio()
{
  char command[G__ONELINE];

  sprintf(command,"stderr=%ld",G__serr);
  G__getexpr(command);
  sprintf(command,"stdout=%ld",G__sout);
  G__getexpr(command);
  sprintf(command,"stdin=%ld",G__sin);
  G__getexpr(command);

}

/******************************************************************
* G__redirectoutput
******************************************************************/
static void G__redirectoutput(com,psout,pserr,psin,asemicolumn)
char *com;
FILE **psout;
FILE **pserr;
FILE **psin;
int asemicolumn;
{
  char *semicolumn;
  char *redirect;
  char *redirectin;
  char *singlequote;
  char *doublequote;
  char *paren;
  char *blacket;
  int issemicolumn;
  char *openmode;
  char filename[G__MAXFILENAME];
  int i=0;
  int j=1;
  int mode; /* 0:stdout, 1:stderr */

  redirect=strrchr(com,'>');
  redirectin=strrchr(com,'<');
  semicolumn=strrchr(com,';');
  singlequote=strrchr(com,'\'');
  doublequote=strrchr(com,'"');
  paren=strrchr(com,')');
  blacket=strrchr(com,']');

  if(0==asemicolumn ||
     (semicolumn && semicolumn<redirect &&
      ((char*)NULL==singlequote||semicolumn>singlequote) &&
      ((char*)NULL==doublequote||semicolumn>doublequote))) {

    if(redirect && 
#ifndef G__OLDIMPLEMENTATION493
       (isspace(*(redirect-1)) ||
	(*(redirect-1)=='2' && isspace(*(redirect-2))) ||
	(*(redirect-1)=='>' && isspace(*(redirect-2))) ||
	(*(redirect-1)=='>' && *(redirect-2)=='2' && isspace(*(redirect-3))))&&
#endif
       ((char*)NULL==singlequote||redirect>singlequote) &&
       ((char*)NULL==doublequote||redirect>doublequote) &&
       ((char*)NULL==paren||redirect>paren) &&
       ((char*)NULL==blacket||redirect>blacket)) {

      /* get filename to redirect 
       *    cint> command > filename
       *                  ^ -------- */
      while((*(redirect+j))) {
	if(!isspace(*(redirect+j))) {
	  filename[i++] = *(redirect+j);
	}
	else if(i) break;
	++j;
      }
      filename[i] = '\0';

      /* check if append mode
       *   cint> command >> filename 
       *                 ^^          */
      --redirect;
      if('>'==(*redirect)) {
	--redirect;
	openmode="a";
      }
      else {
	openmode="w";
      }

      /* check if stderr 
       *   cint> command 2> filename 
       *                 ^^          */
      if('2'==(*redirect)) {
	--redirect;
	mode=G__NUM_STDERR; /* stderr */
      }
      else {
	mode=G__NUM_STDOUT; /* stdout */
      }

      /* check if there is a space char right before redirection
       *   cint> command > filename
       *                ^          */
      if(isspace(*redirect) && filename[0]) {
	/* cut command string at redirect command
	 *   cint> command > filename
	 *                ^0          */
	*redirect='\0';
	/* open redirect file */
#ifdef G__REDIRECTIO
	switch(mode) {
	case G__NUM_STDOUT: /* stdout */
	  *psout = G__sout;
#ifndef G__WIN32
	  if (!strlen(stdoutsav)) strcpy(stdoutsav,ttyname(STDOUT_FILENO));
#endif
	  G__sout = freopen(filename,openmode,G__sout);
	  break;
	case G__NUM_STDERR: /* stderr */
	  *pserr = G__serr;
#ifndef G__WIN32
	  if (!strlen(stderrsav)) strcpy(stderrsav,ttyname(STDERR_FILENO));
#endif
	  G__serr = freopen(filename,openmode,G__serr);
	  break;
	}
#else
	switch(mode) {
	case G__NUM_STDOUT: /* stdout */
	  *psout = G__sout;
	  G__sout = fopen(filename,openmode);
	  break;
	case G__NUM_STDERR: /* stderr */
	  *pserr = G__serr;
	  G__serr = fopen(filename,openmode);
	  break;
	}
	G__update_stdio(); /* update stdout,stderr,stdin in interpreter */
#endif
      } /* isspace */
    } /* redirect && ... */
  }

#ifdef G__REDIRECTIO
  if(0==asemicolumn ||
     (semicolumn && semicolumn<redirectin &&
      ((char*)NULL==singlequote||semicolumn>singlequote) &&
      ((char*)NULL==doublequote||semicolumn>doublequote))) {
    if(redirectin &&
#ifndef G__OLDIMPLEMENTATION493
       isspace(*(redirectin-1)) &&
#endif
       ((char*)NULL==singlequote||redirectin>singlequote) &&
       ((char*)NULL==doublequote||redirectin>doublequote) &&
       ((char*)NULL==paren||redirectin>paren) &&
       ((char*)NULL==blacket||redirectin>blacket)) {
      /* get filename to redirect
       *    cint> command < filename
       *                  ^ -------- */
      i=0;
      j=1;
      while((*(redirectin+j))) {
	if(!isspace(*(redirectin+j))) {
	  filename[i++] = *(redirectin+j);
	}
	else if(i) break;
	++j;
      }
      filename[i] = '\0';

      *redirectin = '\0';
      --redirectin;
      if(isspace(*redirectin)) {
	*psin = G__sin;
#ifndef G__WIN32
	if (!strlen(stdinsav)) strcpy(stdinsav,ttyname(STDIN_FILENO));
#endif
	G__sin = freopen(filename,"r",G__sin);
      }
    }
  }
#endif
}

/******************************************************************
* G__unredirectoutput
******************************************************************/
static void G__unredirectoutput(sout,serr,sin)
FILE **sout;
FILE **serr;
FILE **sin;
{
#ifdef G__REDIRECTIO
  if(*sout) {
#ifdef G__WIN32
    G__sout = freopen("CONOUT$","w",G__sout);
#else
#ifndef G__OLDIMPLEMENTATION495
    G__sout = freopen(stdoutsav,"w",G__sout);
#else
    G__sout = freopen(ttyname(G__NUM_STDOUT),"w",G__sout);
#endif
#endif
    *sout = (FILE*)NULL;
  }
  if(*serr) {
#ifdef G__WIN32
    G__serr = freopen("CONOUT$","w",G__serr);
#else
#ifndef G__OLDIMPLEMENTATION495
    G__serr = freopen(stderrsav,"w",G__serr);
#else
    G__serr = freopen(ttyname(G__NUM_STDERR),"w",G__serr);
#endif
#endif
    *serr = (FILE*)NULL;
  }
  if(*sin) {
#ifdef G__WIN32
    *sin = freopen("CONIN$","r",*sin);
#else
#ifndef G__OLDIMPLEMENTATION495
    *sin = freopen(stdinsav,"r",*sin);
#else
    *sin = freopen(ttyname(G__NUM_STDIN),"r",*sin);
#endif
#endif
    *sin = (FILE*)NULL;
  }
#else
  int flag=0;
  if(*sout) {
    fclose(G__sout);
    G__sout = *sout;
    ++flag;
  }
  if(*serr) {
    fclose(G__serr);
    G__serr = *serr;
    ++flag;
  }
  if(*sin) {
    fclose(G__sin);
    G__sin = *sin;
    ++flag;
  }
  if(flag) G__update_stdio();
#endif
}
#endif

/******************************************************************
* int G__process_cmd()
******************************************************************/
int G__process_cmd(line, prompt, more)
char *line;
char *prompt;
int  *more;
{
  FILE *tempfp;   /* used for input dump file */
  char command[G__ONELINE];
  char syscom[G__ONELINE];
  char editor[64];
  int temp,temp1,temp2;
  int index = -1;
  int ignore=G__PAUSE_NORMAL;
  short double_quote,single_quote;
#ifdef G__ASM
  G__ALLOC_ASMENV;
#endif
  char store_var_type;

/* pass to parent otherwise not re-entrant */
#ifdef G__TMPFILE
  static char tname[G__MAXFILENAME];
#else
  static char tname[L_tmpnam];
#endif
  static struct G__input_file ftemp;

  /* struct G__ifunc_table *ifunc; */

  struct G__var_array *local;

  char *com,*string;
  char *stringb;
  int line_number,filenum;

  G__value buf;

  FILE *G__temp;

  char *evalbase;
  /* void *evalp; */
  int base /* ,digit */ ,num;
  char evalresult[G__ONELINE];
#ifndef G__OLDIMPLEMENTATION464
  FILE* store_stderr=NULL;
  FILE* store_stdout=NULL;
  FILE* store_stdin=NULL;
#endif


#ifndef G__OLDIMPLEMENTATION685
  if(strlen(line)>G__ONELINE-5) {
    fprintf(G__serr,"!!! User command too long !!!\n");
    return(0);
  }
#endif

  *prompt = '\0';
#ifdef G__OLDIMPLEMENTATION463
  if(!fout) fout=G__sout;
#endif

  
  if (com = getenv("EDITOR"))
    strcpy(editor, com);
  else
    strcpy(editor, "vi");
  
  /*** begin ***/
  
  temp = 0;
  while (isspace(line[temp])) ++temp;
  
  if (*more == 0) {
    strcpy(command, line+temp);
    com = strchr(command,'\n');
    if (com) *com = '\0';
    else {
      com = strchr(command,'\r');
      if (com) *com = '\0';
    }
    
    /* #ifdef G__ROOT */
    if(G__rootmode) {
      if (command[0] == '.') {
	strcpy(syscom, command+1);
	strcpy(command, syscom);
      } 
      else if (strcmp(command, "?") == 0) {
	strcpy(command, "help");
      } 
      else if ('\0'!=command[0] && command[0] != '{') {
#ifndef G__OLDIMPLEMENTATION464
	G__redirectoutput(command,&store_stdout,&store_stderr,&store_stdin,1);
#endif
	temp = strlen(command);
	while (isspace(command[temp])) --temp;
	if (command[temp] == ';')
	  sprintf(syscom, "{%s}", command);
	else
	  sprintf(syscom, "{%s;}", command);
	strcpy(command, syscom);
      }
    }
    /* #endif */
  } else {
    sprintf(command, "{%s", line+temp);
    com = strchr(command,'\n');
    if (com) *com = '\0';
    else {
      com = strchr(command,'\r');
      if (com) *com = '\0';
    }
  }
  
  com=command;
  temp=0;
  if(isalpha(command[temp])) 
    while(isalpha(command[temp])) ++temp;
  else if(!isdigit(command[temp]))
    ++temp;
  string=command+temp;

  stringb=string;
  while(isspace(*stringb)&&(*stringb)) ++stringb;
  
  /************************************
   * cint>    X  argument
   *          ^^
   *          com string
   ************************************/
#ifndef G__OLDIMPLEMENTATION464
  if(strlen(com)>=3&&
     0!=strncmp(com,"{",1)&&0!=strncmp(com,"p",1)&&
     0!=strncmp(com,">",1)&&0!=strncmp(com,"<",1)&&
     0!=strncmp(com,"2>",2)) {
    G__redirectoutput(com,&store_stdout,&store_stderr,&store_stdin,0);
  }
#endif
  
  /************************************
   * Break prompt command parser
   ************************************/
  
    if(strncmp("class",com,3)==0) {
      G__more_pause((FILE*)NULL);
      G__display_class(G__sout,string,0);
    }
    else if(strncmp("Class",com,3)==0) {
      G__more_pause((FILE*)NULL);
      G__display_class(G__sout,string,1);
    }

    else if(strncmp("typedef",com,3)==0) {
      G__more_pause((FILE*)NULL);
      G__display_typedef(G__sout,string);
    }

    else if(strncmp("template",com,4)==0) {
      G__more_pause((FILE*)NULL);
      G__display_template(G__sout,string);
    }

    else if(strncmp("macro",com,3)==0) {
      G__more_pause((FILE*)NULL);
      G__display_macro(G__sout,string);
    }

    else if(strncmp("include",com,4)==0) {
      if(0==(*stringb)) G__display_includepath(G__sout);
      else {
	string = stringb+strlen(stringb)-1;
	while(isspace(*string)) *(string--) = 0;
	G__add_ipath(stringb);
      }
    }

    else if(strncmp("file",com,4)==0) {
      G__more_pause((FILE*)NULL);
      G__display_files(G__sout);
    }

    else if(strncmp("string",com,4)==0) {
      G__more_pause((FILE*)NULL);
      G__display_string(G__sout);
    }

    else if(strncmp("garbage",com,4)==0) {
      G__more_pause((FILE*)NULL);
      G__disp_garbagecollection(G__sout);
    }
    else if(strncmp("Garbage",com,4)==0) {
      G__more_pause((FILE*)NULL);
      G__garbagecollection();
      G__disp_garbagecollection(G__sout);
    }

    else if(strncmp("trace",com,3)==0) {
      G__set_tracemode(string);
    }

    else if(strncmp("deltrace",com,4)==0) {
      G__del_tracemode(string);
    }

    else if(strncmp("break",com,4)==0) {
      G__set_classbreak(string);
    }

    else if(strncmp("delbreak",com,4)==0) {
      G__del_classbreak(string);
    }

#ifdef G__SECURITY
#ifndef G__OLDIMPLEMENTATION432
    else if(strncmp("where",com,4)==0) {
      fprintf(G__sout,"FILE:%s LINE:%d interpreter=%s\n"
	      ,view.name,view.line_number,G__nam);
    }

    else if(strncmp("security",com,4)==0) {
      fprintf(G__sout,"security ");
      switch(G__security) {
      case 0x0000000f: fprintf(G__sout,"level0"); break;
      case 0x000000ff: fprintf(G__sout,"level1"); break;
      case 0x00000fff: fprintf(G__sout,"level2"); break;
      case 0x0000ffff: fprintf(G__sout,"level3"); break;
      case 0x000fffff: fprintf(G__sout,"level4"); break;
      case 0x00ffffff: fprintf(G__sout,"level5"); break;
      case 0x0fffffff: fprintf(G__sout,"level6"); break;
      case 0xffffffff: fprintf(G__sout,"level7"); break;
      default: break;
      }
      fprintf(G__sout,"  0x%x\n",G__security);
    }
#else
    else if(strncmp("security",com,4)==0) {
      fprintf(G__sout,"Security flag = 0x%x\n",G__security);
    }
#endif
#endif

    else if(strncmp("scratch",com,4)==0) {
      if(!G__isfilebusy(0)) G__scratch_all();
    }

#ifndef G__ROOT
    else if(strncmp(".",com,1)==0) {
      G__rootmode ^= 1;
      fprintf(G__sout,"!!!Debugger Command mode switched as follows!!!\n");
      if(G__rootmode) {
	fprintf(G__sout,"    > .[command]\n");
	fprintf(G__sout,"    > [statement]\n");
      }
      else {
	fprintf(G__sout,"    > [command]\n");
	fprintf(G__sout,"    > { [statement] }\n");
      }
    }
#endif


    else if(strncmp("reset",com,4)==0) {
#ifdef G__ROOT
      fprintf(G__serr,"!!! Sorry, can not reset interpreter !!!\n");
#else
      if(!G__isfilebusy(0)) { 
	G__scratch_all(); /* To make memory leak check work */
	if(G__commandline[0]) G__init_cint(G__commandline);
	else                  G__init_cint("cint");
      }
      else {
	fprintf(G__serr,"!!! Sorry, can not reset interpreter !!!\n");
      }
#endif
    }

#ifndef G__FONS4
    else if(strncmp("function",com,4)==0) {
#else
    else if(strncmp("function",com,4)==0 || strncmp("L",com,1)==0) {
#endif

      /*******************************************************
       * List up interpreted functions and shared library
       *******************************************************/
      G__more_pause((FILE*)NULL);
      G__listfunc(G__sout,G__PUBLIC_PROTECTED_PRIVATE,(char*)NULL
		    ,(struct G__ifunc_table*)NULL);
      G__listshlfunc(G__sout);
    }

    else if(strncmp("interactive",com,5)==0) {
      G__interactive^=1;
      fprintf(G__sout,"interactive mode=%d\n",G__interactive);
    }

    else if(strncmp("coverage",com,5)==0) {
      temp=0;
      while(isspace(string[temp])) temp++;
      tempfp=fopen(string+temp,"w");
      if(tempfp) {
	fprintf(G__sout,"saving trace coverage to %s\n",string+temp);
	G__dump_tracecoverage(tempfp);
      }
      else {
	fprintf(G__sout,"can not open file %s\n",string+temp);
      }
      fclose(tempfp);
    }

    else if(strncmp("debug",com,5)==0) {
#ifdef G__ASM_DBG
      G__asm_dbg ^= 1;
      fprintf(G__sout,"G__asm_dbg=%d\n",G__asm_dbg);
#endif
    }

    else if(strncmp("asmstep",com,5)==0) {
#ifdef G__ASM_DBG
      G__asm_step ^= 1;
      fprintf(G__sout,"G__asm_step=%d\n",G__asm_step);
#endif
    }

    else if(strncmp("dasm",com,4)==0) {
#ifdef G__ASM_DBG
      G__dasm(G__sout);
#endif
    }

    else if(strncmp("stack",com,4)==0) {
#ifdef G__ASM_DBG
      /* G__display_stack(); */
#endif
    }

    else if(strncmp("status",com,4)==0) {
#ifdef G__ASM_DBG
      fprintf(G__sout,"G__asm_noverflow=%d\n",G__asm_noverflow);
      fprintf(G__sout,"G__no_exec_compile=%d\n",G__no_exec_compile);
      fprintf(G__sout,"G__no_exec=%d\n",G__no_exec);
      fprintf(G__sout,"G__asm_exec=%d\n",G__asm_exec);
      fprintf(G__sout,"G__asm_loopcompile=%d\n",G__asm_loopcompile);
      fprintf(G__sout,"G__asm_cp=%d\n",G__asm_cp);
      fprintf(G__sout,"G__asm_dt=%d\n",G__asm_dt);
#endif
    }

    else if(strncmp("OPTIMIZE",com,1)==0) {
      G__asm_loopcompile = G__int(G__calc(string));
    }
#ifdef G__ASM_WHOLEFUNC
    else if(strncmp("WHOLEFUNC",com,3)==0) {
      G__asm_wholefunction = G__int(G__calc(string));
    }
#endif

    /* interactive return of undefined symbol */
    else if(strncmp("return",com,1)==0) {
#ifndef G__OLDIMPLEMENTATION630
      if(G__interactive_undefined) {
	G__interactivereturnvalue = G__calc(string);
	G__pause_return=1;
	G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
	return(ignore=G__PAUSE_NORMAL);
      }
      else {
	fprintf(G__serr,"!!! Use 'return' command at your own risk !!!\n");
	G__interactivereturnvalue = G__calc(string);
	G__return=G__RETURN_IMMEDIATE;
	G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
	return(ignore=2);
      }
#else
      G__interactivereturnvalue = G__calc(string);
      G__pause_return=1;
      G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
      return(ignore=G__PAUSE_NORMAL);
#endif /* ON630 */
    }

    else if(strncmp(">",com,1)==0) {
      if(strncmp(">>",com,2)==0) index=2;
      else                       index=1;
      while(isspace(command[index])&&command[index]!='\0') index++;
      if((*(command+index))) {
	if(G__sout!=G__stdout) {
	  fprintf(G__stdout,"Old save file closed\n");
	  fclose(G__sout);
	}
	if(strncmp(">>",com,2)==0) G__sout=fopen(command+index,"a");
	else                       G__sout=fopen(command+index,"w");
	if(G__sout) {
	  fprintf(G__stdout,"Output will be saved in file %s! ('>' to display on screen)\n" 
		 ,command+index);
	}
	else {
	  G__sout = G__stdout;
	  fprintf(G__stdout,"Can not open file %s\n",command+index);
	}
      }
      else {
	if(G__sout && G__sout!=G__stdout) fclose(G__sout);
	G__sout = G__stdout;
	fprintf(G__stdout,"Output will be displayed on screen!\n");
      }
    }

    else if(strncmp("2>",com,2)==0) {
      if(strncmp("2>>",com,3)==0) index=3;
      else                        index=2;
      while(isspace(command[index])&&command[index]!='\0') index++;
      if((*(command+index))) {
	if(strncmp("2>>",com,3)==0) G__serr=fopen(command+index,"a");
	else                        G__serr=fopen(command+index,"w");
	if(G__serr) {
	  fprintf(G__stdout,"Error will be saved in file %s! ('2>' to display on screen)\n" 
		 ,command+index);
	}
	else {
	  G__serr = G__stderr;
	  fprintf(G__stdout,"Can not open file %s\n",command+index);
	}
      }
      else {
	if(G__serr && G__serr!=G__stderr) fclose(G__serr);
	G__serr = G__stderr;
	fprintf(G__sout,"Error will be displayed on screen!\n");
      }
    }


    else if(strncmp("<",com,1)==0) {
      /*******************************************************
       * Execute dumpped readline file
       *******************************************************/
      index=1;
      while(isspace(command[index])&&command[index]!='\0') index++;
      tempfp=fopen(command+index,"r");
      if(tempfp) {
	fprintf(G__sout,"Execute readline dumpfile %s\n"
		,command+index);
	G__pushdumpinput(tempfp,1);
      }
      else {
	fprintf(G__sout,"Can not open file %s\n",command+1);
      }
    }

#ifndef G__FONS4
    else if (strncmp("X",com,1)==0) {
#else
    else if (strncmp("x",com,1)==0) {
#endif
      /*******************************************************
       * Execute C/C++ source file. Filename minus extension
       * must match a function in the source file. This function
       * will be automatically executed.
       *******************************************************/ 
#ifndef G__FONS85
      char *com1;
#endif
      temp=0;
      while(isspace(string[temp])) temp++;
#ifndef G__FONS4
      if (string[temp] == '\0') {
        fprintf(G__serr,"Error: no file specified\n");
#ifndef G__OLDIMPLEMENTATION464
	G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
        return(ignore);
      }
#endif
      com = strrchr(string+temp, '(');
      if (com) *com = '\0';
      temp2 = G__prerun;
      G__prerun = 1;  /* suppress warning message if file already loaded */
      temp1 = G__loadfile(string+temp);
      if (temp1 == 1) {
#ifndef G__FONS20
	 G__prerun = 0;
#endif 
         G__unloadfile(string+temp);
         if (G__loadfile(string+temp)) {
            G__prerun = temp2;
	    G__pause_return=1;
#ifndef G__OLDIMPLEMENTATION464
	    G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
            return(ignore);
         }
      }
#ifdef G__OLDIMPLEMENTATION487
#ifdef G__AUTOCOMPILE
      /*************************************************************
       * if '#pragma compile' appears in source code.
       *************************************************************/
      if(G__fpautocc) G__autocc();
#endif
#endif

      G__prerun = temp2;
#ifndef G__FONS85
      com1= strrchr(string+temp,'.');
      if(com) *com='(';
      com = com1;
#else
      if (com) *com = '(';
      com = strrchr(string+temp, '.');
#endif
      if (com) {
#ifndef G__FONS20
	 char *s;
         *com = '\0';
	 s = strrchr(string+temp,'/');
	 if(!s) s=strrchr(string+temp,'\\');
	 if(!s) s=string+temp;
	 else   s++;
	 strcpy(syscom, s);
	 string = strchr(com+1, '(');
	 if (string)
	   strcat(syscom, string);
	 else
	   strcat(syscom, "()");
#else
         *com = '\0';
	 com = strrchr(string+temp,'/');
	 if(!com) strrchr(string+temp,'\\');
	 if(!com) com=string+temp;
	 else     com++;
         sprintf(syscom,"%s()",com);
#endif
#ifndef G__FONS4
         buf=G__calc(syscom);
         G__in_pause=1;
         G__valuemonitor(buf,syscom);
         G__in_pause=0;
         fprintf(G__sout,"%s\n",syscom);
#ifdef G__SECURITY
	 G__security_recover(G__sout);
#endif
#else
         G__calc(syscom);
#endif
      } else {
         fprintf(G__sout, "Expecting . in filename\n");
      }
    }

#ifndef G__FONS4
    else if (strncmp("L",com,1)==0) {
      /*******************************************************
       * Load(Re-Load) a C/C++ source file.
       *******************************************************/

      temp=0;
      while(isspace(string[temp])) temp++;
      if (string[temp] == '\0') {
         fprintf(G__serr,"Error: no file specified\n");
#ifndef G__OLDIMPLEMENTATION464
	 G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
         return(ignore);
      }
      temp2 = G__prerun;
      G__prerun = 1;  /* suppress warning message if file already loaded */
      temp1 = G__loadfile(string+temp);
      if (temp1 == 1) {
         G__unloadfile(string+temp);
         if (G__loadfile(string+temp)) {
            G__prerun = temp2;
            G__pause_return=1;
#ifndef G__OLDIMPLEMENTATION464
	    G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
            return(ignore);
         }
      }
#ifdef G__OLDIMPLEMENTATION487
#ifdef G__AUTOCOMPILE
      /*************************************************************
       * if '#pragma compile' appears in source code.
       *************************************************************/
      if(G__fpautocc) G__autocc();
#endif
#endif
      G__prerun = temp2;
#ifdef G__SECURITY
      G__security_recover(G__sout);
#endif
    }

    else if (strncmp("U",com,1)==0) {
      /*******************************************************
       * Load a C/C++ source file.
       *******************************************************/
      temp=0;
      while(isspace(string[temp])) temp++;
      if (string[temp] == '\0') {
         fprintf(G__serr,"Error: no file specified\n");
#ifndef G__OLDIMPLEMENTATION464
	 G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
         return(ignore);
      }
      G__unloadfile(string+temp);
#ifdef G__SECURITY
      G__security_recover(G__sout);
#endif
    }
#endif

    else if(strncmp("n",com,1)==0) {
      /*******************************************************
       * Create New readline file and start dump.
       *******************************************************/
      index=1;
      while(isspace(command[index])&&command[index]!='\0') index++;
      tempfp=fopen(command+index,"w");
      if(tempfp) {
	fprintf(G__sout
		,"Create new readline dumpfile %s and start dump\n"
		,command+index);
	G__pushdumpinput(tempfp,0);
      }
      else {
	fprintf(G__sout,"Can not open file %s\n",command+1);
      }
    }

    else if(strncmp("y",com,1)==0) {
      /*******************************************************
       * Append readline input to file.
       *******************************************************/
      index=1;
      while(isspace(command[index])&&command[index]!='\0') index++;
      tempfp=fopen(command+index,"a");
      if(tempfp) {
	fprintf(G__sout
		,"Append readline dump to %s\n"
		,command+index);
	G__pushdumpinput(tempfp,0);
      }
      else {
	fprintf(G__sout,"Can not open file %s\n",command+1);
      }
    }

    else if(strncmp("z",com,1)==0) {
      /*******************************************************
       * Stop readline dump.
       *******************************************************/
      if(G__dumpreadline[0]) {
	fprintf(G__sout,"Stop readline dump.\n");
	fclose(G__dumpreadline[0]);
	G__popdumpinput();
      }
      else {
	fprintf(G__sout,"No readline dumpfile in the stack\n");
      }
    }

    else if(strncmp("help",com,1)==0) {
      /*******************************************************
       * Help for break prompt command
       *******************************************************/
      G__more_pause((FILE*)NULL);
#ifndef G__SMALLOBJECT
#ifdef G__ROOT
      G__more(G__sout,"CINT/ROOT C/C++ interpreter interface.\n");
#else
      G__more(G__sout,"cint (C/C++ interpreter) debugger usage:\n");
#endif
      if(G__rootmode) {
	G__more(G__sout,"All commands must be preceded by a . (dot), except\n");
	G__more(G__sout,"for the evaluation statement { } and the ?.\n");
	G__more(G__sout,"===========================================================================\n");
      }
      G__more(G__sout,"Dump:        n [file]  : create new readline dumpfile and start dump\n");
      G__more(G__sout,"             y [file]  : append readline dump to [file]\n");
      G__more(G__sout,"             z         : stop readline dump\n");
      G__more(G__sout,"             < [file]  : execute readline dumpfile\n");
      G__more(G__sout,"             > [file]  : output redirect to [file]\n");
      G__more(G__sout,"             2> [file] : error redirect to [file]\n");
#ifndef G__ROOT
      G__more(G__sout,"             .         : switch command input mode\n");
#endif
#ifdef G__ROOT
      G__more(G__sout,"Help:        ?         : help\n");
      G__more(G__sout,"             help      : help\n");
#else
      G__more(G__sout,"Help:        help      : help\n");
#endif
      G__more(G__sout,"             /[keyword] : help information for keyword\n");
      G__more(G__sout,"Completion:  [nam][Tab] : complete symbol name start with [nam]\n");
      G__more(G__sout,"             [nam][Tab][Tab] : list up all symbol name start with [nam]]\n");
      G__more(G__sout,"Shell:       ![shell]  : execute shell command\n");
      G__more(G__sout,"Source:      v <[line]>: view source code <around [line]>\n");
      G__more(G__sout,"             V [stack] : view source code in function call stack\n");
      G__more(G__sout,"             t         : show function call stack\n");
      G__more(G__sout,"             f [file]  : view source file [file]\n");
      G__more(G__sout,"             T         : turn on trace mode for all source\n");
      G__more(G__sout,"             trace <classname> : turn on trace mode for class\n");
      G__more(G__sout,"             deltrace <classname> : turn off trace mode for class\n");
      G__more(G__sout,"             break [classname] : break at the memberfunc\n");
      G__more(G__sout,"             delbreak [classname] : turn off memberfunc break\n");
      /* G__more(G__sout,"             D         : toggle source code display mode\n"); */
      G__more(G__sout,"Evaluation:  p [expr]  : evaluate expression (no declaration/loop/condition)\n");
      G__more(G__sout,"Evaluation:  s [expr]  : step into expression (no declaration/loop/condition)\n");
      G__more(G__sout,"Evaluation:  S [expr]  : step over expression (no declaration/loop/condition)\n");
      G__more(G__sout,"             {[statements]} : evaluate statement (any kind)\n");
#ifndef G__FONS3
      G__more(G__sout,"             x [file]  : load [file] and evaluate {statements} in the file\n");
      G__more(G__sout,"             X [file]  : load [file] and execute function [file](wo extension)\n");
#endif
      G__more(G__sout,"             E <[file]>: open editor and evaluate {statements} in the file\n");
#ifndef G__FONS4
      G__more(G__sout,"Load/Unload: L [file]  : load [file]\n");
      G__more(G__sout,"             U [file]  : unload [file]\n");
#endif
#ifndef G__FONS2
      G__more(G__sout,"             reset     : reset interpreter environment\n");
#endif
      G__more(G__sout,"Monitor:     g <[var]> : list global variable\n");
      G__more(G__sout,"             l <[var]> : list local variable\n");
#ifndef G__OLDIMPLEMENTATION553
      G__more(G__sout,"             proto <[scope]::>[func] : show function prototype\n");
#endif
      G__more(G__sout,"             class <[name]> : show class definition (one level)\n");
      G__more(G__sout,"             Class <[name]> : show class definition (all level)\n");
      G__more(G__sout,"             typedef <name> : show typedefs\n");
      G__more(G__sout,"             function  : show interpreted functions\n");
      G__more(G__sout,"             macro     : show macro functions\n");
      G__more(G__sout,"             template  : show templates\n");
      G__more(G__sout,"             include   : show include paths\n");
      G__more(G__sout,"             file      : show loaded files\n");
#ifndef G__OLDIMPLEMENTATION432
      G__more(G__sout,"             where     : show current file position\n");
      G__more(G__sout,"             security  : show security level\n");
#endif
      G__more(G__sout,"             garbage   : show garbage collection buffer\n");
      G__more(G__sout,"             Garbage   : Do garbage collection\n");
      G__more(G__sout,"             cover [file] : save trace coverage\n");
      G__more(G__sout,"             return [val] : return undefined symbol value\n");
      G__more(G__sout,"Run:         S         : step over function/loop\n");
      G__more(G__sout,"             s         : step into function/loop\n");
      G__more(G__sout,"             i         : ignore and step over\n");
      G__more(G__sout,"             c <[line]>: continue <to [line]>\n");
      G__more(G__sout,"             e         : step out from function\n");
      G__more(G__sout,"             b [line]  : set break point\n");
      G__more(G__sout,"             db [line] : delete break point\n");
      G__more(G__sout,"             a [assert] : break only if assertion is true\n");
#ifdef G__ASM_DBG
      G__more(G__sout,"             debug   : pcode trace on/off\n");
      G__more(G__sout,"             asmstep : pcode step on/off\n");
      G__more(G__sout,"             status  : show pcode exec status\n");
      G__more(G__sout,"             dasm    : disassembler\n");
#endif
#endif
      G__more(G__sout,"Quit:        q         : quit cint execution\n");
      G__more(G__sout,"             qqq       : quit anyway\n");
    }

    else if(strncmp("/",com,1)==0) {
      G__more_pause((FILE*)NULL);
      /*******************************************************
       * Display keyword help information
       *******************************************************/
      do {
	G__tmpnam(tname);
	G__temp=fopen(tname,"w");
      } while((FILE*)NULL==G__temp && G__setTMPDIR(tname));
      if(G__temp) {
	G__cintrevision(G__temp);
	G__list_sut(G__temp);
	/* compiled and interpreterd objects */
	G__display_class(G__temp,"",0);
	G__display_typedef(G__temp,"");
	G__display_string(G__temp);
	G__display_template(G__temp," ");
	G__display_macro(G__temp,"");
	G__display_files(G__temp);
	G__listfunc(G__temp,G__PUBLIC_PROTECTED_PRIVATE,(char*)NULL
		    ,(struct G__ifunc_table*)NULL);
	G__varmonitor(G__temp,&G__global,"","",0);
	fclose(G__temp);
      
	if(command[strlen(command)-1]==' ') command[strlen(command)-1]='\0';
#ifndef G__OLDIMPLEMENTATION577
	G__display_keyword(G__sout,command+1,tname);
#else
	sprintf(syscom,"grep '%s' %s",command+1,tname);
	system(syscom);
#endif
	remove(tname);
      }
      else {
	fprintf(G__serr,"Error: Tempfile G__temp can not open\n");
      }
    }

    else if(strncmp("!",com,1)==0) {
      /*******************************************************
       * Execute shell command
       *******************************************************/
      system(string);
    }


    else if(strncmp("a",com,1)==0) {
      /*******************************************************
       * Set assertion for break
       *******************************************************/
      command[0]=' ';
      if(command[1]=='\0') {
	sprintf(G__assertion,"");
	fprintf(G__sout,"Break assertion is deleted\n");
      }
      else {
	sprintf(G__assertion,"%s",command+1);
	fprintf(G__sout,"Break only if (%s) is true\n"
		,G__assertion);
      }
    }

#ifndef G__FONS23
    else if(strncmp("g",com,1)==0 || strncmp("G",com,1)==0 ||
            strncmp("l",com,1)==0) {
#else
    else if(strncmp("g",com,1)==0 || strncmp("l",com,1)==0) {
#endif
      /*******************************************************
       * Monitor global variables
       * Monitor local variables
       *******************************************************/
      G__more_pause((FILE*)NULL);
      index=1;
      while(isspace(command[index])) index++;
      temp=index;
      while(command[temp]&&(!isspace(command[temp]))) temp++;
      command[temp]='\0';

#ifdef G__ASM
      G__STORE_ASMENV;
#endif
      store_var_type = G__var_type;
      G__var_type = 'p';

      G__SET_TEMPENV;
      if('g'==command[0]) G__varmonitor(G__sout,&G__global,command+index,"",0);
#ifndef G__FONS23
      else if('G'==command[0]) G__varmonitor(G__sout,&G__global,command+index,"",0);
#endif
      else                G__varmonitor(G__sout,G__p_local,command+index,"",0);
      G__RESET_TEMPENV;

#ifdef G__ASM
      G__RECOVER_ASMENV;
#endif
      G__var_type = store_var_type;
    }

    else if(strncmp("t",com,1)==0) {
      /*******************************************************
       * show function call stack
       *******************************************************/
      G__showstack(G__sout);
    }

    else if(strncmp("T",com,1)==0) {
      /*******************************************************
       * Toggle trace mode
       *******************************************************/
      if(G__istrace==0) {
	G__istrace=1;
#ifdef G__ASM_DBG
	if(string[0]) G__istrace = atoi(string);
#endif
	fprintf(G__sout,"\nTrace mode on %d\n",G__istrace);
      }
      else {
	fprintf(G__sout,"\nTrace mode off\n");
	G__istrace=0;
      }
      G__debug = G__istrace;
      G__setdebugcond();
    }

    else if(strncmp("D",com,1)==0) {
      /*******************************************************
       * Toggle display mode
       *******************************************************/
      if(G__breakdisp==0) {
	G__breakdisp=1;
	fprintf(G__sout,"\nDisplay mode on\n");
      }
      else {
	fprintf(G__sout,"\nDisplay mode off\n");
	G__breakdisp=0;
      }
    }

    else if(strncmp("db",com,2)==0) {
      /*******************************************************
       * Delete break point
       *******************************************************/
      if(1<G__findposition(string,view,&line_number,&filenum)) {
	G__srcfile[filenum].breakpoint[line_number] &= G__NOBREAK;
	fprintf(G__sout,"Break point line %d %s deleted\n"
		,line_number,G__srcfile[filenum].filename);
	G__step=0;
	G__charstep=0;
	G__setdebugcond();
      }
      else {
	fprintf(G__serr,"can not determine where to delete break point\n");
      }
    }

    else if(strncmp("break",com,1)==0) {
      /*******************************************************
       * Set break point
       *******************************************************/
      if(1<G__findposition(string,view,&line_number,&filenum) &&
	 G__srcfile[filenum].breakpoint && G__srcfile[filenum].maxline) {
	fprintf(G__sout,"Break point set to line %d %s\n"
		,line_number,G__srcfile[filenum].filename);
	G__srcfile[filenum].breakpoint[line_number] |= G__BREAK;
	G__step=0;
	G__charstep=0;
	G__setdebugcond();
      }
      else if(0==G__srcfile[filenum].maxline) {
	if((FILE*)NULL==G__srcfile[filenum].fp) {
	  fprintf(G__serr ,"Can not put break point in included file\n");
	}
	else {
	  fprintf(G__serr ,"Setting break point suspended\n");
	  temp=0;
	  while(isspace(string[temp])) ++temp;
	  if('\0'==string[temp]) {
	    sprintf(G__breakline,"%d",G__ifile.line_number);
	    strcpy(G__breakfile,G__srcfile[G__ifile.filenum].filename);
	  }
	  else {
	    strcpy(G__breakline,string+temp);
	  }
	}
      }
      else {
	fprintf(G__serr,"Can not determine where to put break point\n");
      }
    }

    else if(strncmp("continue",com,1)==0) {
      /*******************************************************
       * Continue
       *******************************************************/
      G__stepover=0;
      G__step=0;
      G__charstep=0;
      G__steptrace=0;
      G__setdebugcond();
      if(strlen(string)==0) {
	G__pause_return=1;
#ifndef G__OLDIMPLEMENTATION464
	G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
	return(ignore);
      }
      else if(1<G__findposition(string,view,&line_number,&filenum)) {
	G__srcfile[filenum].breakpoint[line_number] |= G__CONTUNTIL;
	G__pause_return=1;
#ifndef G__OLDIMPLEMENTATION464
	G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
	return(ignore);
      }
      else {
	fprintf(G__serr,"can not determine where to stop\n");
      }
    }

    else if(strncmp("f",com,1)==0) {
      strcpy(syscom,string);
      strcpy(command,syscom);
      /*******************************************************
       * Set view file
       *******************************************************/
      for(temp=0;temp<G__nfile;temp++) {
	if(strcmp(string,G__srcfile[temp].filename)==0) break;
      }
      if(temp>=G__nfile) {
	fprintf(G__serr,"filename %s not loaded\n",string);
      }
      else {
	view.filenum=temp;
	view.fp=G__srcfile[temp].fp;
	strcpy(view.name,G__srcfile[temp].filename);
	view.line_number=1;
	G__pr(G__sout,view);
      }
    }

    else if(strncmp("+",com,1)==0 || strncmp("-",com,1)==0) {
      temp=atoi(command);
      sprintf(command,"%d",view.line_number + temp);
      goto vcommand;
    }

#ifndef G__OLDIMPLEMENTATION553
    else if(strncmp("proto",com,3)==0) G__display_proto(G__sout,string);
#endif

    else if(strncmp("view",com,1)==0) {
      strcpy(syscom,string);
      strcpy(command,syscom);
    vcommand:
      /*******************************************************
       * Display source code
       *******************************************************/
      if(0<G__findposition(command,view,&line_number,&filenum)&&filenum>=0) {
	view.filenum = filenum;
	view.fp = G__srcfile[filenum].fp;
	strcpy(view.name,G__srcfile[filenum].filename);
	view.line_number = line_number;
	G__pr(G__sout,view);
      }
      else {
	fprintf(G__sout,"No proper file view. Can not display source! Use 'f [file]' command\n");
      }
    }

    else if(strncmp("View",com,1)==0) {
      /*******************************************************
       * show source code in stack
       *******************************************************/
      temp1=atoi(command+1);
      temp=0;
      local=G__p_local;
      while(local && temp<temp1-1) {
	++temp;
	local=local->prev_local;
      }
      if(0==temp1) {
	view = G__ifile;
	view_local = G__p_local;
	view_struct_offset=G__store_struct_offset;
	view_tagnum=G__tagnum;
	view_exec_memberfunc=G__exec_memberfunc;
	G__pr(G__sout,view);
      }
      else if(local && local->prev_local) {
	view.filenum = local->prev_filenum ;
	strcpy(view.name,G__srcfile[view.filenum].filename);
	view.fp = G__srcfile[view.filenum].fp;
	view.line_number = local->prev_line_number;
	view_local = local->prev_local;
	view_struct_offset=view_local->struct_offset;
	view_tagnum=view_local->tagnum;
	view_exec_memberfunc=view_local->exec_memberfunc;
	G__pr(G__sout,view);
      }
      else {
	fprintf(G__sout,"Stack isn't that deep\n");
      }
    }


    else if(strncmp("ignore",com,1)==0) {
      /*******************************************************
       * Ignore current statement and step to the next
       *******************************************************/
      G__stepover=0;
      ignore=G__PAUSE_IGNORE;
      G__pause_return=1;
#ifndef G__OLDIMPLEMENTATION464
      G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
      return(ignore);
    }

    else if(strncmp("S",com,1)==0 && 0==(*stringb)) {
      /*****************************************************
       * Step over
       *****************************************************/
      G__stepover=3;
      ignore=G__PAUSE_STEPOVER;
      G__step=1;
      G__charstep=0;
      G__setdebugcond();
      G__pause_return=1;
#ifndef G__OLDIMPLEMENTATION464
      G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
      return(ignore);
    }

    else if(strncmp("s",com,1)==0 && 0==(*stringb)) {
      /*****************************************************
       * Step into
       *****************************************************/
      G__stepover=0;
      ignore=G__PAUSE_NORMAL;
      G__step=1;
      G__charstep=0;
      G__setdebugcond();
      G__pause_return=1;
#ifndef G__OLDIMPLEMENTATION464
      G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
      return(ignore);
    }

    else if(strncmp("e",com,1)==0) {
      /*****************************************************
       * Step out 
       *****************************************************/
      G__stepover=0;
      G__step=0;
      G__charstep=0;
      G__break_exit_func=1;
      G__setdebugcond();
      G__pause_return=1;
#ifndef G__OLDIMPLEMENTATION464
      G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
      return(ignore);
    }

    else if( strncmp(command,"qqq",3)==0 ||
	     strncmp(command,"QQQ",3)==0 ) {
      fprintf(G__sout,"*** Process will be killed ***\n");
      strcpy(command,G__input("Are you sure(y/n)? "));
      if(tolower(command[0])=='y') {
#ifndef G__OLDIMPLEMENTATION464
	G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
	G__close_inputfiles();
	exit(EXIT_SUCCESS); 
      }
      else {
	G__pause_return=1;
#ifndef G__OLDIMPLEMENTATION464
	G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
	return(0);
      }
    }
    
    else if(strncmp("q",com,1)==0) {
      if(G__doingconstruction) {
	fprintf(G__serr,"Use 'qqq' when you quit in the middle of object construction (%d)\n"
		,G__doingconstruction);
	G__pause_return=1;
#ifndef G__OLDIMPLEMENTATION464
	G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
	return(0);
      }

      G__stepover=0;
      G__step=0;
      G__break=0;
      G__steptrace=0;
      G__setdebugcond();

      /*******************************************************
       * Quit session and stop process
       *******************************************************/
      if(G__othermain) {
        /*******************************************************
         * QQ : exit all interpretation
	 * Q  : exit current interpretation
	 *******************************************************/
	if(command[1]=='q') {
	  G__return=G__RETURN_EXIT2;
	  fprintf(G__sout,"Exit current command file.\n");
	}
	else {
	  G__return=G__RETURN_EXIT1;
#ifndef G__ROOT
	  fprintf(G__sout,"Exit current interpretation.\n");
#endif
	}
	/*****************************************
	* return 2 so that user can utilize 
	* while(G__pause()==0) ;
	*****************************************/
	G__pause_return=1;
#ifndef G__OLDIMPLEMENTATION464
	G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
	return(ignore=2);
      }
      else {
        /*******************************************************
	 * Q  : exit process
	 *******************************************************/
#ifndef G__OLDIMPLEMENTATION464
	G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
#ifdef G__DEBUG
	 G__exit(EXIT_SUCCESS);
#else
	 G__close_inputfiles();
	 exit(EXIT_SUCCESS); 
#endif
      }
    }


#ifndef G__FONS3
    else if(strncmp("E",com,1)==0 || strncmp("x",com,1)==0) {
#else
    else if(strncmp("E",com,1)==0) {
#endif
      /*******************************************************
       * Open tempfile to edit 
       *******************************************************/
      temp=1;
      while(isspace(command[temp])) temp++;
#ifndef G__FONS3
      if (com[0] == 'E') {
#endif
	if(command[temp]=='\0') {
	  if('\0'==G__tempc[0]) G__tmpnam(G__tempc);
	  sprintf(syscom,"%s %s",editor, G__tempc);
	  system(syscom);
	  sprintf(syscom,G__tempc);
	}
	else {
	  sprintf(syscom,"%s %s",editor, command+temp);
	  system(syscom);
	  sprintf(syscom,command+temp);
	}
#ifndef G__FONS3
      } else {
	if (command[temp]=='\0') {
	  fprintf(G__serr,"Error: no file specified\n");
#ifndef G__OLDIMPLEMENTATION464
	  G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
	  return(ignore);
	}
	sprintf(syscom,command+temp);
      }
#endif

      /*******************************************************
       * Execute temp file
       *******************************************************/
      G__SET_TEMPENV;
      buf=G__exec_tempfile(syscom);
#ifndef G__OLDIMPLEMENTATION437
      G__security = G__srcfile[G__ifile.filenum].security;
#endif
#ifndef G__ROOT
      G__in_pause=1;
      G__valuemonitor(buf,syscom);
      G__in_pause=0;
      if(buf.type) fprintf(G__sout,"%s\n",syscom);
#endif
      G__RESET_TEMPENV;

#ifdef G__SECURITY
      G__security_recover(G__sout);
#endif

      if(G__return!=G__RETURN_NON) {
	G__pause_return=1;
#ifndef G__OLDIMPLEMENTATION464
	G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
	return(ignore);
      }
      
    }

    else if(strncmp("{",com,1)==0) {
      /*******************************************************
       * Evaluate sequencial statements
       *******************************************************/
      if (*more == 0) {
	do {
	  G__tmpnam(tname);
	  ftemp.fp = fopen(tname,"w");
	} while((FILE*)NULL==ftemp.fp && G__setTMPDIR(tname));
      } 
      else {
	com = command+1;
      }

      if(!ftemp.fp) {
	fprintf(G__serr,"Error: could not create file %s\n",tname);
      }
      else {
	temp = *more; 
        temp1 = 0;
	double_quote=0; single_quote=0;
	while(com[temp1]!='\0') {
	  switch(com[temp1]) {
	  case '"':
	    if(single_quote==0) {
		    double_quote ^= 1;
	    }
	    break;
	  case '\'':
	    if(double_quote==0) {
		    single_quote ^= 1;
	    }
	    break;
	  case '{':
	    if((single_quote==0)&&(double_quote==0))
	      temp++;
	    break;
	  case '}':
	    if((single_quote==0)&&(double_quote==0))
	      temp--;
	    break;
	  }
	  temp1++;
	}
	if (temp>0) {
	  fprintf(ftemp.fp,"%s\n",com);
	  strcpy(prompt,"end with '}'> ");
          *more = temp;
#ifndef G__FONS23
	  G__pause_return=0;
#else
	  G__pause_return=1;
#endif
          return(0);
	} else {
	  string = strrchr(com,'}');
	  if(string) {
#ifndef G__OLDIMPLEMENTATION464
	    G__redirectoutput(string
			      ,&store_stdout,&store_stderr,&store_stdin,0);
#endif
	    *string = '\0';
	  }
	  fprintf(ftemp.fp,"%s\n",com);
	  fprintf(ftemp.fp,"}\n");
          *more = 0;
          prompt[0] = '\0';
        }
	fclose(ftemp.fp);

	/*******************************************************
	 * Execute temp file
	 *******************************************************/
	G__more_pause((FILE*)NULL);
	G__SET_TEMPENV;
	buf=G__exec_tempfile(tname);
	remove(tname);
	G__in_pause=1;
	G__valuemonitor(buf,syscom);
	G__in_pause=0;
	if(buf.type) fprintf(G__sout,"%s\n",syscom);
	G__RESET_TEMPENV;
#ifdef G__SECURITY
	G__security_recover(G__sout);
#endif

	if(G__return!=G__RETURN_NON) {
	  G__pause_return=1;
#ifndef G__OLDIMPLEMENTATION464
	  G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
	  return(ignore);
	}
	
      }
    }

#ifndef G__OLDIMPLEMENTATION696
    else if(strncmp("p",com,1)==0 || 
	    strncmp("s",com,1)==0 || strncmp("S",com,1)==0) {
      G__STORE_EVALENV;
#else
    else if(strncmp("p",com,1)==0) {
#endif
      /*******************************************************
       * Evaluate statement
       *******************************************************/
      G__more_pause((FILE*)NULL);
      strcpy(syscom,string);
      strcpy(command,syscom);
      if(strlen(command)>0) {

#ifndef G__OLDIMPLEMENTATION464
	G__redirectoutput(command,&store_stdout,&store_stderr,&store_stdin,1);
#endif
	
#ifdef G__ASM
	G__STORE_ASMENV;
#endif
	store_var_type = G__var_type;
	G__var_type='p';
	
	single_quote=0;
	double_quote=0;
	evalbase=command;
	temp=1;
	while(temp) {
	  evalbase++;
	  switch(*evalbase) {
	  case '"':
	    if(single_quote==0) {
	      double_quote ^= 1;
	    }
	    break;
	  case '\'':
	    if(double_quote==0) {
	      single_quote ^= 1;
	    }
	    break;
	  case '\\':
	    if(double_quote==0&&single_quote==0) {
	      temp=0;
	    }
	    else {
	      evalbase++;
	    }
	    break;
	  case '\0':
	    if(double_quote==0&&single_quote==0) {
	      temp=0;
	      evalbase=(char *)NULL;
	    }
	    break;
	  }
	}
	G__SET_TEMPENV;
	if(evalbase) {
	  *evalbase='\0';
	  temp1=1;
	  num=0;
	  if(isdigit(evalbase[temp1])) {
	    while(isdigit(evalbase[temp1])) {
	      syscom[temp1-1]=evalbase[temp1];
	      ++temp1;
	    }
	    syscom[temp1-1]='\0';
	    num=atoi(syscom);
	  }
	  switch(tolower(evalbase[temp1])) {
	  case 'x':
	  case 'h':
	    base=16;
	    break;
	  case 'b':
	  case 'z':
	    base=2;
	    break;
	  case 'o':
	    base=8;
	    break;
	  case 'd':
	    base=10;
	    break;
	  case 'n':
	    base=0;
	    break;
	  }
	  if(num<=0) {
	    buf = G__calc(command);
	    G__in_pause=1;
	    if(base==0) {
	      G__valuemonitor(buf,syscom);
	    }
	    else {
	      G__getbase(buf.obj.i ,base ,0,evalresult);
	      sprintf(syscom,"0%c%s" ,evalbase[temp1] ,evalresult);
	    }
	    G__in_pause=0;
	    fprintf(G__sout,"%s\n",syscom);
	  }
	  else {
	    for(temp=0;temp<num;temp++) {
	      if(temp%2==0) {
		sprintf(syscom,"&%s+%d" ,command+1 ,temp);
		buf=G__calc(syscom);
		fprintf(G__sout,"\n0x%x: " ,buf.obj.i);
	      }
	      sprintf(syscom,"*(&%s+%d)" ,command+1 ,temp);
	      buf = G__calc(syscom);
	      G__in_pause=1;
	      if(base==0) {
		G__valuemonitor(buf,syscom);
	      }
	      else {
		G__getbase(buf.obj.i,base ,0,evalresult);
		sprintf(syscom,"0%c%s" ,evalbase[temp1] ,evalresult);
	      }
	      G__in_pause=0;
	      fprintf(G__sout,"%30s ",syscom);
	    }
	    fprintf(G__sout,"\n");
	  }
	}
	else {
	  buf = G__calc(command);
#ifndef G__OLDIMPLEMENTATION696
	  G__step=store_step;
#endif
	  G__in_pause=1;
	  G__valuemonitor(buf,syscom);
	  G__in_pause=0;
	  if(buf.type) fprintf(G__sout,"%s\n",syscom);
	  if('u'==buf.type && buf.obj.i) {
#ifndef G__OLDIMPLEMENTATION444
	    G__objectmonitor(G__sout,buf.obj.i,buf.tagnum,"");
#else
	    G__incsetup_memvar(buf.tagnum);
	    G__varmonitor(G__sout
			  ,G__struct.memvar[buf.tagnum],"","  ",buf.obj.i);
#endif
	  }
	}
	G__RESET_TEMPENV;
	
#ifdef G__ASM
	G__RECOVER_ASMENV;
#endif
	G__var_type = store_var_type;
	
	/*****************************************************
	 * free temp object buffer
	 *****************************************************/
	G__free_tempobject();

#ifdef G__SECURITY
	G__security_recover(G__sout);
#endif
	
	if(G__return!=G__RETURN_NON) {
	  G__pause_return=1;
#ifndef G__OLDIMPLEMENTATION464
	  G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
#ifndef G__OLDIMPLEMENTATION696
	  G__RESTORE_EVALENV;
#endif
	  return(ignore);
	}
      }
      else {
	ignore=G__stepover;
	G__pause_return=1;
#ifndef G__OLDIMPLEMENTATION464
	G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
#ifndef G__OLDIMPLEMENTATION696
	G__RESTORE_EVALENV;
#endif
	return(ignore);
      }
#ifndef G__OLDIMPLEMENTATION696
      G__RESTORE_EVALENV; 
#endif
    }

    else if(strcmp(com,"P")==0) { /* Dummy pause when reading from dump file */
      ignore=G__PAUSE_NORMAL;
      G__pause_return=0;
#ifndef G__OLDIMPLEMENTATION464
      G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
      return(ignore);
    }

    else if(strcmp(com,"")==0) {
      ignore=G__stepover;
      G__pause_return=1;
#ifndef G__OLDIMPLEMENTATION464
      G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
      return(ignore);
    }

    else {
      fprintf(G__serr,"Unknown interpreter command '%s'\n",com);
    }

#ifndef G__OLDIMPLEMENTATION464
    G__unredirectoutput(&store_stdout,&store_stderr,&store_stdin);
#endif
  return(ignore); /* never happens, avoiding lint error */
}



/**************************************************************************
* G__setaccess
*
*
**************************************************************************/
int G__setaccess(statement,iout)
char *statement;
int iout;
{
  if(7==iout && strcmp(statement,"public:")==0) {
    G__access=G__PUBLIC;
  }
  else if(10==iout && strcmp(statement,"protected:")==0) {
    G__access=G__PROTECTED;
  }
  else if(8==iout && strcmp(statement,"private:")==0) {
    G__access=G__PRIVATE;
  }
}


/*
 * Local Variables:
 * c-tab-always-indent:nil
 * c-indent-level:2
 * c-continued-statement-offset:2
 * c-brace-offset:-2
 * c-brace-imaginary-offset:0
 * c-argdecl-indent:0
 * c-label-offset:-2
 * compile-command:"make -k"
 * End:
 */
