/* /% C %/ */
/***********************************************************************
 * cint (C/C++ interpreter)
 ************************************************************************
 * Source file tmplt.c
 ************************************************************************
 * Description:
 *  Class and member function template
 ************************************************************************
 * Copyright(c) 1991~1996  [hp] Hewlett Packard Japan, All rights reserved
 * Author                  Masaharu Goto (gotom@jpn.hp.com)
 * Copyright(c) 1995~1997  Masaharu Goto (MXJ02154@niftyserve.or.jp)
 *
 * Permission to use, copy, modify and distribute this software and its 
 * documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear
 * in supporting documentation.  Hewlett-Packard Japan Company makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 ************************************************************************/

#include "common.h"

#ifndef G__OLDIMPLEMENTATION691
/***********************************************************************
* G__IntList_init()
***********************************************************************/
void G__IntList_init(body,iin,prev)
struct G__IntList *body;
long iin;
struct G__IntList *prev;
{
  body->i=iin;
  body->next=(struct G__IntList*)NULL;
  body->prev = prev;
}

/***********************************************************************
* G__IntList_new()
***********************************************************************/
struct G__IntList* G__IntList_new(iin,prev)
long iin;
struct G__IntList *prev;
{
  struct G__IntList *body;
  body = malloc(sizeof(struct G__IntList));
  G__IntList_init(body,iin,prev);
  return(body);
}

/***********************************************************************
* G__IntList_add()
***********************************************************************/
void G__IntList_add(body,iin)
struct G__IntList *body;
long iin;
{
  while(body->next) body=body->next;
  body->next = G__IntList_new(iin,body);
}

/***********************************************************************
* G__IntList_addunique()
***********************************************************************/
void G__IntList_addunique(body,iin)
struct G__IntList *body;
long iin;
{
  while(body->next) {
    if(body->i==iin) return;
    body=body->next;
  }
  if(body->i==iin) return;
  body->next = G__IntList_new(iin,body);
}

/***********************************************************************
* G__IntList_delete(body)
***********************************************************************/
void G__IntList_delete(body)
struct G__IntList *body;
{
  if(body->prev && body->next) {
    body->prev->next = body->next;
    body->next->prev = body->prev;
  }
  else if(body->next) {
    body->next->prev = (struct G__IntList*)NULL;
  }
  else if(body->prev) {
    body->prev->next = (struct G__IntList*)NULL;
  }
  free(body);
}

/***********************************************************************
* G__IntList_find()
***********************************************************************/
struct G__IntList* G__IntList_find(body,iin)
struct G__IntList *body;
long iin;
{
  while(body->next) {
    if(body->i == iin) return(body);
    body=body->next;
  }
  if(body->i == iin) return(body);
  return((struct G__IntList*)NULL);
}

/***********************************************************************
* G__IntList_free()
***********************************************************************/
void G__IntList_free(body)
struct G__IntList *body;
{
  if(!body) return;
  if(body->prev) body->prev->next = (struct G__IntList*)NULL;
  while(body->next) G__IntList_free(body->next);
  free(body);
}

/***********************************************************************
* G__instantiate_templateclasslater()
*
*  instantiation of forward declared template class body
***********************************************************************/
void G__instantiate_templateclasslater(deftmpclass)
struct G__Definedtemplateclass *deftmpclass;
{
  /* forward declaration of template -> instantiation -> 
   * definition of template NOW instantiate forward declaration */
  struct G__IntList *ilist = deftmpclass->instantiatedtagnum;
  int store_def_tagnum=G__def_tagnum;
  int store_tagdefining=G__tagdefining;
  int store_def_struct_member=G__def_struct_member;
  char tagname[G__ONELINE];
  while(ilist) {
    G__ASSERT(ilist->i>=0);
    strcpy(tagname,G__struct.name[ilist->i]);
    if(-1!=G__struct.parent_tagnum[ilist->i]) {
      G__def_tagnum=G__struct.parent_tagnum[ilist->i];
      G__tagdefining=G__struct.parent_tagnum[ilist->i];
      G__def_struct_member=1;
    }
    else {
      G__def_tagnum=store_def_tagnum;
      G__tagdefining=store_tagdefining;
      G__def_struct_member=store_def_struct_member;
    }
    G__instantiate_templateclass(tagname);
    ilist = ilist->next;
  }
  G__def_tagnum=store_def_tagnum;
  G__tagdefining=store_tagdefining;
  G__def_struct_member=store_def_struct_member;
}

/***********************************************************************
* G__instantiate_templatememfunclater()
*
*  instantiation of forward declared template class member function
***********************************************************************/
void G__instantiate_templatememfunclater(deftmpclass,deftmpmemfunc)
struct G__Definedtemplateclass *deftmpclass;
struct G__Definedtemplatememfunc *deftmpmemfunc;
{
  struct G__IntList* ilist=deftmpclass->instantiatedtagnum;
  struct G__Charlist call_para;
  char templatename[G__ONELINE];
  char tagname[G__ONELINE];
  char *arg;
  int npara=0;
  int store_def_tagnum=G__def_tagnum;
  int store_tagdefining=G__tagdefining;
  int store_def_struct_member=G__def_struct_member;

  while(ilist) {
    G__ASSERT(0<=ilist->i);
    strcpy(tagname,G__struct.name[ilist->i]);
    strcpy(templatename,tagname);
    arg = strchr(templatename,'<');
    if(arg) {
      *arg='\0';
      ++arg;
    }
    else {
      arg = "";
    }
    call_para.string=(char*)NULL;
    call_para.next = (struct G__Charlist*)NULL;
    G__gettemplatearglist(arg,&call_para,deftmpclass->def_para,&npara);
    if(-1!=G__struct.parent_tagnum[ilist->i]) {
      G__def_tagnum=G__struct.parent_tagnum[ilist->i];
      G__tagdefining=G__struct.parent_tagnum[ilist->i];
      G__def_struct_member=1;
    }
    else {
      G__def_tagnum=store_def_tagnum;
      G__tagdefining=store_tagdefining;
      G__def_struct_member=store_def_struct_member;
    }
    G__replacetemplate(templatename,tagname,&call_para
		       ,deftmpmemfunc->def_fp
		       ,deftmpmemfunc->line
		       ,deftmpmemfunc->filenum
		       ,&(deftmpmemfunc->def_pos)
		       ,deftmpclass->def_para
		       ,0
		       ,npara
		       );
    G__freecharlist(&call_para);
    ilist=ilist->next;
  }
  G__def_tagnum=store_def_tagnum;
  G__tagdefining=store_tagdefining;
  G__def_struct_member=store_def_struct_member;
}
#endif /* ON691 */


#ifdef G__TEMPLATECLASS
/***********************************************************************
* G__cattemplatearg()
*
* Concatinate templatename and template arguments
***********************************************************************/
int G__cattemplatearg(tagname,charlist)
char *tagname;
struct G__Charlist *charlist;
{
  char *p;
  p=strchr(tagname,'<');
  ++p;
  /* B<int,5*2>
   *   ^ => p */
  while(charlist->next) {
    strcpy(p,charlist->string);
    p+=strlen(charlist->string);
    charlist=charlist->next;
    if(charlist->next) {
      *p=','; ++p;
    }
  }
  *p='>'; ++p;
  *p='\0'; ++p;
  return 0;
}

/***********************************************************************
* G__catparam()
*
* Cancatinate parameter string to libp->parameter[0] and return.
*
*  "B<int"   "double"   "5>"     =>    "B<int,double,5>"
***********************************************************************/
char *G__catparam(libp,catn,connect)
struct G__param *libp;
int catn;
char *connect;
{
  int i;
  char *p;
  int lenconnect;
  /* B<int\0
   *      ^ => p */
  p = libp->parameter[0]+strlen(libp->parameter[0]);
  lenconnect = strlen(connect);
  for(i=1;i<catn;i++) {
    strcpy(p,connect);
    p+=lenconnect;
    strcpy(p,libp->parameter[i]);
    p+=strlen(libp->parameter[i]);
  }
  return(libp->parameter[0]);
}

/**************************************************************************
* G__read_formal_templatearg()
*
*  template<class T,class E,int S> ...
*           ^
**************************************************************************/
struct G__Templatearg *G__read_formal_templatearg()
{
  struct G__Templatearg *targ;
  struct G__Templatearg *p;
  char type[G__MAXNAME],name[G__MAXNAME];
  int c;
  int stat=1;

  do {
    
    /* allocate entry of template argument list */
    if(stat) {
      p = (struct G__Templatearg *)malloc(sizeof(struct G__Templatearg));
      p->next = (struct G__Templatearg *)NULL;
      /* store entry of the template argument list */
      targ = p;
      stat=0;
    }
    else {
      p->next = (struct G__Templatearg *)malloc(sizeof(struct G__Templatearg));
      p=p->next;
      p->next = (struct G__Templatearg *)NULL;
    }

    /*  template<class T,class E,int S> ...
     *           ^                            */
    c = G__fgetname(type,"<");
    if(strcmp(type,"class")==0) {
      p->type = G__TMPLT_CLASSARG;
    }
    else if('<'==c && strcmp(type,"template")==0) {
      c=G__fignorestream(">");
      c=G__fgetname(type,"");
      G__ASSERT(0==strcmp(type,"class"));
      p->type = G__TMPLT_TMPLTARG;
    }
    else {
      if(strcmp(type,"int")==0) p->type = G__TMPLT_INTARG;
      else if(strcmp(type,"size_t")==0) p->type = G__TMPLT_SIZEARG;
      else if(strcmp(type,"unsignedint")==0) p->type = G__TMPLT_UINTARG;
      else if(strcmp(type,"unsigned")==0) p->type = G__TMPLT_UINTARG;
      else if(strcmp(type,"char")==0) p->type = G__TMPLT_CHARARG;
      else if(strcmp(type,"unsignedchar")==0) p->type = G__TMPLT_UCHARARG;
      else if(strcmp(type,"short")==0) p->type = G__TMPLT_SHORTARG;
      else if(strcmp(type,"unsignedshort")==0) p->type = G__TMPLT_USHORTARG;
      else if(strcmp(type,"long")==0) p->type = G__TMPLT_LONGARG;
      else if(strcmp(type,"unsignedlong")==0) p->type = G__TMPLT_ULONGARG;
      else if(strcmp(type,"float")==0) p->type = G__TMPLT_FLOATARG;
      else if(strcmp(type,"double")==0) p->type = G__TMPLT_DOUBLEARG;
      else {
	if(G__dispsource) {
	  fprintf(G__serr,"Limitation: template argument type '%s' may cause problem");
	  G__genericerror((char*)NULL);
	}
	p->type = G__TMPLT_INTARG;
      }
    }

    /*  template<class T,class E,int S> ...
     *                 ^                     */
    c = G__fgetstream(name,",>="); /* G__fgetstream_tmplt() ? */
    p->string=(char*)malloc(strlen(name)+1);
    strcpy(p->string,name);

    if('='==c) {
#ifndef G__OLDIMPLEMENTATION663
      c = G__fgetstream_template(name,",>"); /* G__fgetstream_tmplt() ? */
#else
      c = G__fgetstream(name,",>"); /* G__fgetstream_tmplt() ? */
#endif
      p->default_parameter=(char*)malloc(strlen(name)+1);
      strcpy(p->default_parameter,name);
    }
    else {
      p->default_parameter=(char*)NULL;
    }

    /*  template<class T,class E,int S> ...
     *                   ^                  */
  } while(','==c) ;

  /*  template<class T,class E,int S> ...
   *                                 ^    */

  return(targ);
}

#ifdef G__TEMPLATEMEMFUNC
/**************************************************************************
* G__createtemplatememfunc()
*  template<class T,class E,int S> type A<T,E,S>::f() { .... }
*                                         ^    
**************************************************************************/
G__createtemplatememfunc(new_name)
char *new_name;
{
  int c;
  struct G__Definedtemplateclass *deftmpclass;
  struct G__Definedtemplatememfunc *deftmpmemfunc;
  int os=0;

  /* funcname="*f()" "&f()" */
  while('*'==new_name[os] || '&'==new_name[os]) ++os;

  /* get defined tempalte class identity */
  deftmpclass = G__defined_templateclass(new_name+os);
  if(!deftmpclass) {
    /* error */
    fprintf(G__serr,"Error: Template class %s not defined",new_name+os);
    G__genericerror((char*)NULL);
  }
  else {
    /* get to the end of defined member function list */
    deftmpmemfunc = &(deftmpclass->memfunctmplt) ; 
    while(deftmpmemfunc->next) deftmpmemfunc=deftmpmemfunc->next;
    
    /* allocate member function template list */
    deftmpmemfunc->next = (struct G__Definedtemplatememfunc*)malloc(sizeof(struct G__Definedtemplatememfunc));
    deftmpmemfunc->next->next = (struct G__Definedtemplatememfunc*)NULL;
    
    /* set file position */
    deftmpmemfunc->def_fp = G__ifile.fp;
    deftmpmemfunc->line = G__ifile.line_number;
    deftmpmemfunc->filenum = G__ifile.filenum;
    fgetpos(G__ifile.fp,&deftmpmemfunc->def_pos);

#ifndef G__OLDIMPLEMENTATION691
    /* if member function is defined after template class instantiation 
     * instantiate member functions here */
    if(deftmpclass->instantiatedtagnum) {
      G__instantiate_templatememfunclater(deftmpclass,deftmpmemfunc);
    }
#endif
  }
}
#endif

/**************************************************************************
* G__createtemplateclass()
*  template<class T,class E,int S> class A { .... };
*                                 ^    
**************************************************************************/
int G__createtemplateclass(new_name,targ
#ifndef G__OLDIMPLEMENTATION691
                          ,isforwarddecl
#endif
                          )
char *new_name;
struct G__Templatearg *targ;
#ifndef G__OLDIMPLEMENTATION691
int isforwarddecl;
#endif
{
  struct G__Definedtemplateclass *deftmpclass;
  int hash,i;
  int override=0;
#ifndef G__OLDIMPLEMENTATION682
  int env_tagnum = G__get_envtagnum();
#endif
  
  /* Search for the end of list */
  deftmpclass = &G__definedtemplateclass;
  G__hash(new_name,hash,i)
  while(deftmpclass->next) {
    if(deftmpclass->hash==hash && strcmp(deftmpclass->name,new_name)==0
#ifndef G__OLDIMPLEMENTATION682
       && env_tagnum==deftmpclass->parent_tagnum
#endif
       ) {
#ifndef G__OLDIMPLEMENTATION691
      if(0==deftmpclass->isforwarddecl && deftmpclass->def_fp) {
#else
      if(deftmpclass->def_fp) {
#endif
	fprintf(G__serr,"Error: template %s duplicate definition");
	G__genericerror(NULL);
      }
      override=1;
      break;
    }
    deftmpclass=deftmpclass->next;
  }
  
  if(!override) {
    /* store name and hash key */
    deftmpclass->name = malloc(strlen(new_name)+1);
    strcpy(deftmpclass->name,new_name);
    deftmpclass->hash=hash;
  }
  
#ifndef G__OLDIMPLEMENTATION682
  /* store parent_tagnum */
  deftmpclass->parent_tagnum = env_tagnum;
#endif

  /* store template argument list */
#ifndef G__OLDIMPLEMENTATION691
  if(!override || !deftmpclass->def_para) deftmpclass->def_para=targ;
  else G__freetemplatearg(targ);
#else
  deftmpclass->def_para=targ;
#endif
  
  /* store file pointer, line number and position */
  deftmpclass->def_fp = G__ifile.fp;
  if(G__ifile.fp) fgetpos(G__ifile.fp,&deftmpclass->def_pos);
  deftmpclass->line = G__ifile.line_number;
  deftmpclass->filenum = G__ifile.filenum;
  
  if(!override) {
    /* allocate and initialize next list */
    deftmpclass->next = (struct G__Definedtemplateclass*)malloc(sizeof(struct G__Definedtemplateclass));
    deftmpclass->next->def_para = (struct G__Templatearg*)NULL;
    deftmpclass->next->next = (struct G__Definedtemplateclass*)NULL;
    deftmpclass->next->name = (char*)NULL;
    deftmpclass->next->hash = 0;
    deftmpclass->next->memfunctmplt.next 
      = (struct G__Definedtemplatememfunc*)NULL;
    deftmpclass->next->def_fp = (FILE*)NULL;
#ifndef G__OLDIMPLEMENTATION691
    deftmpclass->next->isforwarddecl = 0;
    deftmpclass->next->instantiatedtagnum = (struct G__IntList*)NULL;
#endif
  }
  /* skip template class body */
  if(targ) G__fignorestream(";");
  /*  template<class T,class E,int S> class A { .... };
   *                                                   ^ */

#ifndef G__OLDIMPLEMENTATION691
  /* forward declaration of template -> instantiation -> 
   * definition of template NOW instantiate forward declaration */
  if(1==deftmpclass->isforwarddecl && 0==isforwarddecl &&
     deftmpclass->instantiatedtagnum) {
    G__instantiate_templateclasslater(deftmpclass);
  }
  deftmpclass->isforwarddecl = isforwarddecl;
#endif

  return(0);
}

/***********************************************************************
* G__defined_templateclass()
*
* Check if the template is declared
* Not used now, but maybe in future I might need this to handle case 4,5
***********************************************************************/
struct G__Definedtemplateclass *G__defined_templateclass(name)
char *name;
{
#ifndef G__OLDIMPLEMENTATION682
  struct G__Definedtemplateclass *deftmplt;
  int hash,temp;
  long dmy_struct_offset=0;
  char atom_name[G__ONELINE];
  int env_tagnum=G__get_envtagnum();
  int scope_tagnum = -1;
  struct G__inheritance *baseclass;

  /* return if no name */
  if('\0'==name[0]||strchr(name,'.')||strchr(name,'-'))
     return((struct G__Definedtemplateclass *)NULL);

  /* get a handle for using declaration info */
#ifndef G__OLDIMPLEMENTATION686
  if(-1!=env_tagnum) baseclass = G__struct.baseclass[env_tagnum];
  else               baseclass = &G__globalusingnamespace;
  if(0==baseclass->basen) baseclass = (struct G__inheritance*)NULL;
#else
  if(-1!=env_tagnum) {
    baseclass = G__struct.baseclass[env_tagnum];
    if(0==baseclass->basen) baseclass = (struct G__inheritance*)NULL;
  }
  else {
    baseclass = (struct G__inheritance*)NULL;
  }
#endif

  /* scope operator resolution, A::templatename<int> ... */
  strcpy(atom_name,name);
  G__hash(atom_name,hash,temp)
  G__scopeoperator(atom_name,&hash,&dmy_struct_offset,&scope_tagnum);

  /* search for template name and scope match */
  deftmplt = &G__definedtemplateclass;
  while(deftmplt) {
    if(hash==deftmplt->hash && strcmp(atom_name,deftmplt->name)==0) {
      /* look for ordinary scope resolution */
      if((-1==scope_tagnum&&(-1==deftmplt->parent_tagnum||
			     env_tagnum==deftmplt->parent_tagnum))||
	 scope_tagnum==deftmplt->parent_tagnum) {
	return(deftmplt);
      }
      else if(-1==scope_tagnum) {
	int env_parent_tagnum = env_tagnum;
	if(baseclass) {
	  /* look for using directive scope resolution */
	  for(temp=0;temp<baseclass->basen;temp++) {
	    if(baseclass->basetagnum[temp]==deftmplt->parent_tagnum) {
	      return(deftmplt);
	    }
	  }
	}
	/* look for enclosing scope resolution */
	while(-1!=env_parent_tagnum) {
	  env_parent_tagnum = G__struct.parent_tagnum[env_parent_tagnum];
	  if(env_parent_tagnum==deftmplt->parent_tagnum) return(deftmplt);
	}
      }
    }
    deftmplt=deftmplt->next;
  }
  return((struct G__Definedtemplateclass *)NULL);

#else
  struct G__Definedtemplateclass *deftmplt;
  int hash,temp;
  if('\0'==name[0]) return((struct G__Definedtemplateclass *)NULL);
  G__hash(name,hash,temp)
  deftmplt = &G__definedtemplateclass;
  while(deftmplt) {
    if(hash==deftmplt->hash&&strcmp(name,deftmplt->name)==0) return(deftmplt);
    deftmplt=deftmplt->next;
  }
  return((struct G__Definedtemplateclass *)NULL);
#endif
}

/***********************************************************************
* G__declare_template()
*
* Entry of template declaration parsing
*
*   template<class T> class A { };
*            ^
*   template<class T> type A <T>::f() { }
*   template<class T> A <T>::B<T> A <T>::f() { }
*   template<class T> A <T>::A() { }
*            ^
*   template<class T> type A <T>::staticmember;
*   template<class T> A <T>::B<T> A <T>::staticmember;
*            ^
*   template<class T> type f() { }
*   template<class T> A <T>::B<T> f() { }
*            ^
***********************************************************************/
void G__declare_template()
{
  char temp[G__ONELINE];
  fpos_t pos;
  int store_line_number;
  struct G__Templatearg *targ;
  int c;
  char *p;
  char temp2[G__ONELINE];
  int ismemvar=0;
#ifndef G__OLDIMPLEMENTATION691
  int isforwarddecl = 0;
#endif

  /* Set a flag that template or macro is included in the source file,
   * so that this file won't be closed even with -cN option after preRUN */
  ++G__macroORtemplateINfile;

  /* read template argument declaration */
  targ=G__read_formal_templatearg();

  /*  template<class T,class E,int S> ...
   *                                 ^   store this position below */
  fgetpos(G__ifile.fp,&pos);
  store_line_number = G__ifile.line_number;
  /* if(G__dispsource) G__disp_mask=1000; */

  do {
    c=G__fgetname_template(temp,"(<");
  } while(strcmp(temp,"inline")==0||strcmp(temp,"const")==0) ;

  /* template class */
  if(strcmp(temp,"class")==0 || strcmp(temp,"struct")==0) {
#ifndef G__OLDIMPLEMENTATION662
    c = G__fgetstream(temp,":{;"); /* read template name */
    if(';'==c) {
#ifndef G__OLDIMPLEMENTATION691
      isforwarddecl = 1;
#else
      /* G__nosupport("Forward declaration of template"); */
      fprintf(G__serr,"Limitation Forward declaration of template ignored");
      G__printlinenum();
      return;
#endif
    }
#else
    c = G__fgetstream(temp,":{"); /* read template name */
#endif
    fsetpos(G__ifile.fp,&pos);
    if(G__dispsource) G__disp_mask=0;
    G__ifile.line_number = store_line_number;
#ifndef G__OLDIMPLEMENTATION691
    G__createtemplateclass(temp,targ,isforwarddecl);
#else
    G__createtemplateclass(temp,targ);
#endif
    return;
  }

  /* Judge between template class member and global function */
  if('<'==c) {
    /* must judge if this is a constructor or other function 
     *1 template<class T> A<T>::f()  constructor
     *2 template<class T> A<T>::B<T> A<T>::f()
     *3 template<class T> A<T> A<T>::f()
     *4 template<class T> A<T>::B<T> f()
     *5 template<class T> A<T> f()
     *6 template<class T> A<T> A<T>::v;
     *                      ^>^            */ 
    c = G__fgetstream_template(temp2,">");
    c = G__fgetname_template(temp2,"(;");
    if(isspace(c)) {
      if(strcmp(temp2,"::~")==0) 
	c = G__fgetname_template(temp2+3,"(;");
      else if(strcmp(temp2,"::")==0)
	c = G__fgetname_template(temp2+2,"(;");
      else if((p=strstr(temp2,"::"))&&strcmp(p,"::operator")==0) {
	/* A<T> A<T>::operator T () { } */
	c='<'; /* this is a flag indicating this is a member function tmplt */
      }
      else if(strcmp(temp2,"operator")==0) {
	c = G__fgetstream(temp2+8,"(");
      }
    }
    if(';'==c) ismemvar=1;
    if('('==c||';'==c) {
      /*1 template<class T> A<T>::f()           ::f
       *3 template<class T> A<T> A<T>::f()      A<T>::f
       *6 template<class T> A<T> A<T>::v;       A<T>::v
       *5 template<class T> A<T> f()            f        */
      p=strchr(temp2,':');
      if(p) {
	c='<';
	if(p!=temp2) {
	  p=strchr(temp2,'<');
	  *p='\0';  /* non constructor/destructor member function */
	  strcpy(temp,temp2);
	}
      }
      else {
	strcpy(temp,temp2); 
     }
    }
    else if('<'==c) {
      /* Do nothing */
    }
    else { /* if(strncmp(temp,"::",2)==0) { */
      /*2 template<class T> A<T>::B<T> A<T>::f()  ::B<T>
       *4 template<class T> A<T>::B<T> f()        ::B<T> */
      c=G__fgetstream(temp,"(;<");
      /* Judge by c? '('  global or '<' member */
    }
    /*
    else {
      p=strchr(temp2,'<');
      if(p) {
	*p = '\0';
	strcpy(temp,temp2);
	c='<';
      }
      else if(isspace(c)&&'\0'==temp[0]) {
	c=G__fgetspace();
      }
    }
    */
  }
  else { /* if('<'==c) */
    /* template<..> inline|const type A<T,S>::f() { ... }
     * template<..> inline|const type f(T a,S b) { ... }
     *                               ^              */
    do {
      c=G__fgetname_template(temp,"(<"); 
    } while('('!=c && '<'!=c) ;
  }

  /* template<..> type A<T,S>::f() { ... }
   * template<..> type f(T a,S b) { ... }
   *                     ^                   */
  if('<'==c && strcmp(temp,"operator")!=0) {
    /* member function template */
    fsetpos(G__ifile.fp,&pos);
    G__ifile.line_number = store_line_number;
    if(G__dispsource) G__disp_mask=0;
    G__createtemplatememfunc(temp);
    /* skip body of member function template */
    c = G__fignorestream("{;");
    if(';'!=c) c = G__fignorestream("}");
    G__freetemplatearg(targ);
  }
  else {
    if(G__dispsource) G__disp_mask=0;
    /* global function template */
    if(strcmp(temp,"operator")==0) {
      /* in case of operator< operator<= operator<< */
      temp[8]=c; /* operator< */
      c=G__fgetstream(temp+9,"(");
    }
    G__createtemplatefunc(temp,targ,store_line_number,&pos);
  }
}

#ifndef G__OLDIMPLEMENtATION461
/**************************************************************************
* G__templatemaptypename()
*
* separate and evaluate template argument list
**************************************************************************/
static void G__templatemaptypename(string)
char *string;
{
#ifdef G__OLDIMPLEMENTATION609_YET
  int typenum;
#endif
#ifndef G__OLDIMPLEMENTATION675
  if(strncmp(string,"const",5)==0 && string[5]!=' ') {
#else
  if(strncmp(string,"const",5)==0) {
#endif
    if(
       strcmp(string+5,"int")==0||
       strcmp(string+5,"unsignedint")==0||
       strcmp(string+5,"char")==0||
       strcmp(string+5,"unsignedchar")==0||
       strcmp(string+5,"short")==0||
       strcmp(string+5,"unsignedshort")==0||
       strcmp(string+5,"long")==0||
       strcmp(string+5,"unsignedlong")==0||
       strcmp(string+5,"double")==0||
       strcmp(string+5,"float")==0||
       strcmp(string+5,"int*")==0||
       strcmp(string+5,"unsignedint*")==0||
       strcmp(string+5,"char*")==0||
       strcmp(string+5,"unsignedchar*")==0||
       strcmp(string+5,"short*")==0||
       strcmp(string+5,"unsignedshort*")==0||
       strcmp(string+5,"long*")==0||
       strcmp(string+5,"unsignedlong*")==0||
       strcmp(string+5,"double*")==0||
       strcmp(string+5,"float*")==0
       || G__istypename(string+5)
      ) {
      int len=strlen(string);
      while(len>=5) {
	string[len+1] = string[len];
	--len;
      }
      string[5] = ' ';
      string += 6;
    }
  }
  if(strcmp(string,"shortint")==0) strcpy(string,"short");
  else if(strcmp(string,"shortint*")==0) strcpy(string,"short*");
  else if(strcmp(string,"longint")==0) strcpy(string,"long");
  else if(strcmp(string,"longint*")==0) strcpy(string,"long*");
  else if(strcmp(string,"unsignedchar")==0) strcpy(string,"unsigned char");
  else if(strcmp(string,"unsignedchar*")==0) strcpy(string,"unsigned char*");
  else if(strcmp(string,"unsignedint")==0) strcpy(string,"unsigned int");
  else if(strcmp(string,"unsignedint*")==0) strcpy(string,"unsigned int*");
  else if(strcmp(string,"unsignedlong")==0||
	  strcmp(string,"unsignedlongint")==0) 
    strcpy(string,"unsigned long");
  else if(strcmp(string,"unsignedlong*")==0||
	  strcmp(string,"unsignedlongint*")==0) 
    strcpy(string,"unsigned long*");
  else if(strcmp(string,"unsignedshort")==0||
	  strcmp(string,"unsignedshortint")==0) 
    strcpy(string,"unsigned short");
  else if(strcmp(string,"unsignedshort*")==0||
	  strcmp(string,"unsignedshortint*")==0) 
    strcpy(string,"unsigned short*");
#ifdef G__OLDIMPLEMENTATION609_YET
  else if(-1!=(typenum=G__defined_typename(string))) {
    if('u'==G__newtype.type[typenum] && -1!=G__newtype.tagnum[typenum]) {
      strcpy(string,G__struct.name[G__newtype.tagnum[typenum]]);
    }
    else if(0==G__newtype.nindex[typenum]) {
      strcpy(string,G__type2string(G__newtype.type[typenum],-1,-1
				   ,G__newtype.reftype[typenum],0));
    }
  } 
#endif
}
#endif

#ifndef G__OLDIMPLEMENTATION664
/**************************************************************************
* G__expand_def_template_arg()   by Scott Snyder 1997/Oct/17
*
* Returns a malloc'd string.
**************************************************************************/
char* G__expand_def_template_arg (str_in, def_para, charlist)
char* str_in;
struct G__Charlist *charlist;
struct G__Templatearg *def_para;
{
  char *punctuation=" \t\n;:=+-)(*&^%$#@!~'\"\\|][}{/?.>,<";
  int siz_out = strlen (str_in) * 2;
  char* str_out;
  char* temp;
  int iout;
  int iin;
  int single_quote;
  int double_quote;
  char c;

  if (siz_out < 10)
    siz_out = 10;

  temp = (char*) malloc (siz_out + 1);

  str_out = (char*) malloc (siz_out + 1);
  iout = 0;

  iin = 0;

  /* The text has been through the reader once, so we shouldn't
     have to worry about comments.
     We should still be prepared to handle quotes though. */

  single_quote = double_quote = 0;

  do {
    int lreslt;
    char* reslt = temp;

    c = G__getstream (str_in, &iin, temp, punctuation);

    if (*reslt != '\0' && 0 == single_quote && 0 == double_quote) {
      struct G__Charlist* cl = charlist;
      struct G__Templatearg* ta = def_para;

      while (cl && cl->string) {
        G__ASSERT (ta && ta->string);
        if (strcmp (ta->string, reslt) == 0) {
          reslt = cl->string;
          break;
        }
        ta = ta->next;
        cl = cl->next;
      }
    }

    /* ??? Does this handle backslash escapes properly? */
    if('\''==c && 0==double_quote) 
      single_quote = single_quote ^ 1 ;
    else if('"'==c && 0==single_quote) 
      double_quote = double_quote ^ 1 ;

    lreslt = strlen (reslt);
    if (iout + lreslt + 1 > siz_out) {
      siz_out = (iout + lreslt + 1) * 2;
      str_out = (char*) realloc (str_out, siz_out + 1);
    }

    strcpy (str_out + iout, reslt);
    iout += lreslt;
    str_out[iout++] = c;
  } while (c != '\0');

  str_out[iout] = '\0';

  free (temp);

  return str_out;
}
#endif /* ON664 */

/**************************************************************************
* G__gettemplatearglist()
*
* separate and evaluate template argument list
**************************************************************************/
#ifndef G__OLDIMPLEMENTATION664
int G__gettemplatearglist(paralist,charlist_in,def_para_in ,pnpara)
char *paralist;
struct G__Charlist *charlist_in;
struct G__Templatearg *def_para_in;
int *pnpara;
#else /* ON664 */
int G__gettemplatearglist(paralist,charlist,def_para ,pnpara)
char *paralist;
struct G__Charlist *charlist;
struct G__Templatearg *def_para;
int *pnpara;
#endif /* ON664 */
{
#ifndef G__OLDIMPLEMENTATION664
  struct G__Charlist *charlist = charlist_in;
  struct G__Templatearg *def_para = def_para_in;
#endif
  int isrc;
  char string[G__ONELINE];
  char temp[G__ONELINE];
  int c;
  G__value buf;
  int searchflag=0;
  
  /**************************************************************
  * explicitly given template argument
  **************************************************************/
  c=',';
  isrc=0;
  while(','==c) {
#ifndef G__OLDIMPLEMENTATION688
    if('\0'==paralist[0]) break;
#endif
    c = G__getstream_template(paralist,&isrc,string,",>\0");
    if(def_para) {
      switch(def_para->type) {
      case G__TMPLT_CLASSARG:
#ifndef G__OLDIMPLEMENTATION608
	strcpy(temp,string);
	G__templatemaptypename(temp);
	if(strcmp(temp,string)!=0) {
	  searchflag=1;
	  strcpy(string,temp);
	}
	break;
#endif
      case G__TMPLT_TMPLTARG:
	break;
      default:
	buf = G__getexpr(string);
	G__string(buf,temp);
	if(strcmp(temp,string)!=0) {
	  searchflag=1;
	  strcpy(string,temp);
	}
	break;
      }
      def_para = def_para->next;
    }
    else {
      G__genericerror("Error: Too many template arguments");
    }
#ifdef G__OLDIMPLEMENTATION608
    G__templatemaptypename(string);
#endif
    charlist->string = malloc(strlen(string)+1);
    strcpy(charlist->string,string);
    charlist->next = (struct G__Charlist*)malloc(sizeof(struct G__Charlist));
    charlist->next->next = (struct G__Charlist *)NULL;
    charlist->next->string = (char *)NULL;
    charlist = charlist->next;
    ++(*pnpara);
  }

  /**************************************************************
  * default template argument
  **************************************************************/
  if(def_para) {
    while(def_para) {
      if(def_para->default_parameter) {
	strcpy(string,def_para->default_parameter);
	G__templatemaptypename(string);
#ifndef G__OLDIMPLEMENTATION664
        charlist->string = G__expand_def_template_arg (string,def_para_in,
                                                       charlist_in);
#else
	charlist->string = malloc(strlen(string)+1);
	strcpy(charlist->string,string);
#endif
	charlist->next=(struct G__Charlist*)malloc(sizeof(struct G__Charlist));
	charlist->next->next = (struct G__Charlist *)NULL;
	charlist->next->string = (char *)NULL;
	charlist = charlist->next;
      }
      else {
	G__genericerror("Error: Too few template arguments");
      }
      def_para=def_para->next;
    }
  }

  return(searchflag);
}

/***********************************************************************
* G__instantiate_templateclass()
*
***********************************************************************/
int G__instantiate_templateclass(tagname)
char *tagname;
{
  int tagnum;
  int hash,temp;
  char templatename[G__ONELINE];
  char *arg;
  struct G__Definedtemplateclass *deftmpclass;
  struct G__Charlist call_para;
#ifdef G__TEMPLATEMEMFUNC
  struct G__Definedtemplatememfunc *deftmpmemfunc;
#endif
  int npara=0;
#ifndef G__OLDIMPLEMENTATION682
  long dmy_struct_offset=0;
  char atom_name[G__ONELINE];
  int env_tagnum=G__get_envtagnum();
  int scope_tagnum = -1;
  struct G__inheritance *baseclass;
#endif /* ON682 */

#ifdef G__ASM
  G__asm_noverflow=0;
#endif

  call_para.string=(char*)NULL;
  call_para.next = (struct G__Charlist*)NULL;

  /* separate template name and argument into templatename and arg  */
  strcpy(templatename,tagname);
  arg = strchr(templatename,'<');
#ifndef G__OLDIMPLEMENTATION688
  if(arg) {
    *arg='\0';
    ++arg;
  }
  else {
    arg = "";
  }
#else
  *arg='\0';
  ++arg;
#endif

#ifndef G__OLDIMPLEMENTATION682
  /* prepare for using directive scope resolution */
#ifndef G__OLDIMPLEMENTATION686
  if(-1!=env_tagnum) baseclass = G__struct.baseclass[env_tagnum];
  else               baseclass = &G__globalusingnamespace;
  if(0==baseclass->basen) baseclass = (struct G__inheritance*)NULL;
#else
  if(-1!=env_tagnum) {
    baseclass = G__struct.baseclass[env_tagnum];
    if(0==baseclass->basen) baseclass = (struct G__inheritance*)NULL;
  }
  else {
    baseclass = (struct G__inheritance*)NULL;
  }
#endif

  /* scope operator resolution, A::templatename<int> ... */
  strcpy(atom_name,templatename);
  G__hash(atom_name,hash,temp)
  G__scopeoperator(atom_name,&hash,&dmy_struct_offset,&scope_tagnum);

  /* search for template class name */
  deftmpclass = &G__definedtemplateclass;
  while(deftmpclass) {
    if(hash==deftmpclass->hash && strcmp(atom_name,deftmpclass->name)==0) {
      /* look for ordinary scope resolution */
      if((-1==scope_tagnum&&(-1==deftmpclass->parent_tagnum||
			     env_tagnum==deftmpclass->parent_tagnum))||
	 scope_tagnum==deftmpclass->parent_tagnum) {
	goto exit_loop;
      }
      else if(-1==scope_tagnum) {
	int env_parent_tagnum = env_tagnum;
	if(baseclass) {
	  /* look for using directive scope resolution */
	  for(temp=0;temp<baseclass->basen;temp++) {
	    if(baseclass->basetagnum[temp]==deftmpclass->parent_tagnum) {
	      goto exit_loop;
	    }
	  }
	}
	/* look for enclosing scope resolution */
	while(-1!=env_parent_tagnum) {
	  env_parent_tagnum = G__struct.parent_tagnum[env_parent_tagnum];
	  if(env_parent_tagnum==deftmpclass->parent_tagnum) goto exit_loop;
	}
      }
    }
    deftmpclass=deftmpclass->next;
  }
 exit_loop:

#else /* ON682 */

  G__hash(templatename,hash,temp);
  /* search for template class name */
  deftmpclass = &G__definedtemplateclass;
  while(deftmpclass) {
    if(hash==deftmpclass->hash && strcmp(templatename,deftmpclass->name)==0) {
      break;
    }
    deftmpclass=deftmpclass->next;
  }

#endif /* ON682 */
 
  /* if no such template, error */
  if(!deftmpclass) {
    fprintf(G__serr,"Error: no such template %s",tagname);
    G__genericerror((char*)NULL);
    return(-1);
  }


  if(!deftmpclass->def_fp) {
    fprintf(G__serr,"Limitation: Can't instantiate precompiled template %s"
	    ,tagname);
    G__genericerror(NULL);
    return(-1);
  }

  /* separate and evaluate template argument */
  if(G__gettemplatearglist(arg,&call_para,deftmpclass->def_para,&npara)) {
    /* If evaluated template argument is not identical as string to
     * the original argument, recursively call G__defined_tagname()
     * to find actual tagname. */
    G__cattemplatearg(tagname,&call_para);
    tagnum = G__defined_tagname(tagname,1);
    G__freecharlist(&call_para);
    return(tagnum);
  }

#ifndef G__OLDIMPLEMENTATION682
  if(-1!=scope_tagnum) {
    int i=0;
    char *p = strrchr(templatename,':');
    while(*p) templatename[i++] = *(++p);
    sprintf(tagname,"%s<%s",templatename,arg);
  }
#endif /* ON682 */

  /* store tagnum */
  tagnum = G__struct.alltag;

  /* string substitution and parse substituted template class definition */
  G__replacetemplate(templatename,tagname,&call_para
		     ,deftmpclass->def_fp
		     ,deftmpclass->line
		     ,deftmpclass->filenum
		     ,&(deftmpclass->def_pos)
		     ,deftmpclass->def_para
#ifndef G__OLDIMPLEMENTATION691
		     ,deftmpclass->isforwarddecl?2:1
#else
		     ,1
#endif
		     ,npara
		     );

#ifdef G__TEMPLATEMEMFUNC
  deftmpmemfunc= &(deftmpclass->memfunctmplt);
  while(deftmpmemfunc->next) {
    G__replacetemplate(templatename,tagname,&call_para
		       ,deftmpmemfunc->def_fp
		       ,deftmpmemfunc->line
		       ,deftmpmemfunc->filenum
		       ,&(deftmpmemfunc->def_pos)
		       ,deftmpclass->def_para
		       ,0
		       ,npara
		       );
    deftmpmemfunc=deftmpmemfunc->next;
  }
#endif /* G__TEMPLATEFUNC */

#ifndef G__OLDIMPLEMENTATION691
  tagnum = G__defined_tagname(tagname,2);
  if(-1!=tagnum) {
    if(deftmpclass->instantiatedtagnum) {
      G__IntList_addunique(deftmpclass->instantiatedtagnum,tagnum);
    }
    else {
      deftmpclass->instantiatedtagnum=G__IntList_new(tagnum,NULL);
    }
  }
#endif /* ON691 */

  /* free template argument lisst */
  G__freecharlist(&call_para);

  /* return instantiated class template id */
  return(tagnum);
}

/**************************************************************************
* G__replacetemplateclass()
*
* Replace template string and prerun
*
**************************************************************************/
void G__replacetemplate(templatename,tagname,callpara
			,def_fp,line,filenum,pdef_pos,def_para,isclasstemplate
			,npara
			)
char *templatename;
char *tagname;
struct G__Charlist *callpara;
FILE *def_fp;
int line;
int filenum;
fpos_t *pdef_pos;
struct G__Templatearg *def_para;
int isclasstemplate;
int npara;
{
  fpos_t store_mfpos;
  int store_mfline;
  fpos_t orig_pos;
  fpos_t pos;
  int c,c2;
  int mparen;
  char symbol[G__ONELINE];
  char *punctuation=" \t\n;:=+-)(*&^%$#@!~'\"\\|][}{/?.>,<";
  int double_quote=0,single_quote=0;
  struct G__input_file store_ifile;
  int store_prerun;
  int store_tagnum,store_def_tagnum;
#ifndef G__OLDIMPLEMENTATION440
  int store_tmplt_def_tagnum;
#endif
  int store_tagdefining,store_def_struct_member;
  int store_var_type;
  int store_breaksignal;
  int store_no_exec_compile;
  int store_asm_noverflow;
  int store_func_now;
  int store_func_page;
  int store_decl;
  struct G__ifunc_table *store_ifunc;
#ifndef G__OLDIMPLEMENTATION521
  int slash=0;
#endif

  /*******************************************************************
   * open macro and template substitution file and get ready for
   * template instantiation
   *******************************************************************/
  
  /* get tmpfile file pinter */
  if(G__mfp==NULL) {
    G__openmfp();
    fgetpos(G__mfp,&G__nextmacro);
    G__mline=1;
    store_mfline=0;
  }
  else {
    fgetpos(G__mfp,&store_mfpos);
    store_mfline=G__mline;
    fsetpos(G__mfp,&G__nextmacro);
  }

  if(G__dispsource) {
    fprintf(G__serr,"\n!!!Instantiating template %s\n",tagname);
  }
  
  /* print out header */
  ++G__mline;
  fprintf(G__mfp,"// template %s  FILE:%s LINE:%d\n"
	  ,tagname ,G__ifile.name,G__ifile.line_number);
  if(G__dispsource) {
    fprintf(G__serr,"// template %s  FILE:%s LINE:%d\n"
	    ,tagname ,G__ifile.name,G__ifile.line_number);
  }
  
  /* store restart position, used later in this function */
#ifndef G__OLDIMPLEMENTATION605
  /* protect the case when template is instantiated from command line */
  if(G__ifile.fp) fgetpos(G__ifile.fp,&orig_pos);
#else
  fgetpos(G__ifile.fp,&orig_pos);
#endif
  fgetpos(G__mfp,&pos);
  
  /* set file pointer and position */
  store_ifile = G__ifile;
  G__ifile.fp = def_fp;
  G__ifile.line_number = line;
  G__ifile.filenum = filenum;
  fsetpos(G__ifile.fp,pdef_pos);

  /* output file position indicator */
  ++G__mline;
  fprintf(G__mfp,"# %d \"%s\"\n"
	  ,G__ifile.line_number,G__srcfile[G__ifile.filenum].filename);
  if(G__dispsource) {
    fprintf(G__serr,"# %d \"%s\"\n"
	    ,G__ifile.line_number,G__srcfile[G__ifile.filenum].filename);
  }
  
  /*******************************************************************
   * read template definition and substitute template arguments
   *******************************************************************/
  
  /* read definition and substitute */
  mparen=0;
  while(1) {
    G__disp_mask = 10000;
    c = G__fgetstream(symbol,punctuation);
    if('\0' != symbol[0]) {
      if(0==double_quote && 0==single_quote) {
	if(isspace(c)) {
	  c2=c;
	  while(isspace(c=G__fgetc())) ;
	  if('<'!=c) {
	    fseek(G__ifile.fp,-1,SEEK_CUR);
	    c=c2;
	  }
	}
	if(G__templatesubstitute(symbol,callpara,def_para,templatename
				 ,tagname,c,npara) && '>'!=c) {
	  char ignorebuf[G__ONELINE];
	  c=G__fgetstream(ignorebuf,">");
	  G__ASSERT('>'==c);
	  c='>';
	}
      }
      fprintf(G__mfp,"%s",symbol);
      if(G__dispsource) fprintf(G__serr,"%s",symbol);
    }

#ifndef G__OLDIMPLEMENTATION521
    if(1==slash) {
      slash=0;
      if('/'==c) {
	G__fignoreline();
	fprintf(G__mfp,"/\n");
	if(G__dispsource) fprintf(G__serr,"/\n");
	++G__mline;
	continue;
      }
      else if('*'==c) {
	fprintf(G__mfp,"/\n");
	if(G__dispsource) fprintf(G__serr,"/\n");
	++G__mline;
	G__skip_comment();
	continue;
      }
    }
#endif
    
    if(0==single_quote && 0==double_quote) {
      if('{'==c) ++mparen;
      else if('}'==c) {
	--mparen;
	if(0==mparen) {
	  fputc(c,G__mfp);
	  if(G__dispsource) fputc(c,G__serr);
	  break;
	}
      }
      else if(';'==c && 0==mparen) break;
    }
    
    if('\''==c && 0==double_quote) 
      single_quote = single_quote ^ 1 ;
    
    else if('"'==c && 0==single_quote) 
      double_quote = double_quote ^ 1 ;

#ifndef G__OLDIMPLEMENTATION521
    if('/'==c) slash=1;
#endif
    
    fputc(c,G__mfp);
    if(G__dispsource) fputc(c,G__serr);
    if('\n'==c||'\r'==c) ++G__mline;
  } 

#ifndef G__OLDIMPLEMENTATION691
  if(2==isclasstemplate) {
    fprintf(G__mfp,";");
    if(G__dispsource) fprintf(G__serr,";");
  }
  else if(1==isclasstemplate) {
#else
  if(isclasstemplate) {
#endif
    G__fgetstream(symbol,";");
    fprintf(G__mfp,"%s ;",symbol);
    if(G__dispsource) fprintf(G__serr,"%s ;",symbol);
  }
  else if(';'==c) {
    fputc(c,G__mfp);
    if(G__dispsource) fputc(c,G__serr);
  }
  fputc('\n',G__mfp);
  if(G__dispsource) fputc('\n',G__serr);
  ++G__mline;
  
  /* finish string substitution */
  G__disp_mask=0;
  fgetpos(G__mfp,&G__nextmacro);
  fflush(G__mfp);

  /*******************************************************************
  * rewind tmpfile and parse template class or function
  ********************************************************************/
  if(G__dispsource) {
    fprintf(G__serr,"!!! Reading template %s\n",tagname);
  }

  fsetpos(G__mfp,&pos);
  G__ifile.fp=G__mfp;

  store_prerun = G__prerun;
  store_tagnum = G__tagnum;
  store_def_tagnum = G__def_tagnum;
  store_tagdefining = G__tagdefining;
  store_tmplt_def_tagnum = G__tmplt_def_tagnum;
  store_def_struct_member = G__def_struct_member;
  store_var_type = G__var_type;
  store_breaksignal=G__breaksignal;
  store_no_exec_compile = G__no_exec_compile;
  store_asm_noverflow = G__asm_noverflow;
  store_func_now=G__func_now;
  store_func_page=G__func_page;
  store_decl=G__decl;
  store_ifunc = G__p_ifunc;

  G__prerun=1;
  G__tagnum = -1;
  G__tmplt_def_tagnum = G__def_tagnum;
#ifndef G__OLDIMPLEMENTATION682
  /* template instantiation only in namespace scope */
  if(-1!=G__tagdefining && 'n'!=G__struct.type[G__tagdefining]) {
    /* even if current scope is not namespace, it may be enclosed by
     * namespace. Check scope hierarchy. */
    int flag=0;
    int parent_tagnum = G__struct.parent_tagnum[G__tagdefining];
    while(-1!=parent_tagnum) {
      if('n'==G__struct.type[parent_tagnum]) {
        /* namespace has found in scope hierarchy. Set scope environment */
        G__def_tagnum = parent_tagnum;
        G__tagdefining = parent_tagnum;
        G__def_struct_member = 1;
        flag=1;
        break;
      }
      parent_tagnum = G__struct.parent_tagnum[parent_tagnum];
    }
    if(0==flag) {
      /* template instantiation in global scope */
      G__def_tagnum = -1;
      G__tagdefining = -1;
      G__def_struct_member = 0;
    }
  }
#else
  /* Do I need to do 'G__def_tagnum = -1' here , or must not ?? maybe YES */
  G__def_tagnum = -1;
  G__tagdefining = -1;
  G__def_struct_member = 0;
#endif
  G__var_type = 'p';
  G__breaksignal=0;
  G__asm_noverflow=0;
  G__no_exec_compile=0;
  G__func_now = -1;
  G__func_page = 0;
  G__decl=0;
  G__p_ifunc = &G__ifunc;
  
  G__exec_statement();

  G__func_now=store_func_now;
  G__func_page=store_func_page;
  G__decl=store_decl;
  G__p_ifunc = store_ifunc;
  G__asm_noverflow=store_asm_noverflow;
  G__no_exec_compile=store_no_exec_compile;
  G__prerun=store_prerun;
  G__tagnum=store_tagnum;
  G__tmplt_def_tagnum = store_tmplt_def_tagnum;
  G__def_tagnum=store_def_tagnum;
  G__tagdefining = store_tagdefining;
  G__def_struct_member = store_def_struct_member;
  G__var_type=store_var_type;
  G__breaksignal=store_breaksignal;
  
  /* restore input file */
  G__ifile = store_ifile;
  G__security = G__srcfile[G__ifile.filenum].security;
#ifndef G__OLDIMPLEMENTATION605
  /* protect the case when template is instantiated from command line */
  if(G__ifile.fp) fsetpos(G__ifile.fp,&orig_pos);
#else
  fsetpos(G__ifile.fp,&orig_pos);
#endif

  if(store_mfline) fsetpos(G__mfp,&store_mfpos);
}

/**************************************************************************
* G__templatesubstitute()
*
* Substitute macro argument
*
**************************************************************************/
int G__templatesubstitute(symbol,callpara,defpara,templatename,tagname,c
			  ,npara)
char *symbol;
struct G__Charlist *callpara;
struct G__Templatearg *defpara;
char *templatename;
char *tagname;
int c;
int npara;
{
  int flag=0;
  static int state=0;

  /* template name substitution */
  if(strcmp(symbol,templatename)==0) {
    if('<'!=c) {
      strcpy(symbol,tagname);
      state=0;
      return(flag);
    }
    else {
      state=1;
      return(flag);
    }
  }

  while(defpara) {
    if(strcmp(defpara->string,symbol)==0) {
      if(callpara->string) {
	strcpy(symbol,callpara->string);
      }
      else if(defpara->default_parameter) {
	strcpy(symbol,defpara->default_parameter);
      }
      else {
	fprintf(G__serr,"Error: template argument for %s missing"
		,defpara->string);
	G__genericerror((char*)NULL);
      }
      if(state) {
	if(state==npara) flag=1;
	++state;
      }
      
      break;
    }
    else {
      state=0;
    }
    defpara = defpara->next;
    if(callpara) callpara=callpara->next;
  }
  return(flag);
}

/**************************************************************************
* G__freedeftemplateclass()
**************************************************************************/
void G__freedeftemplateclass(deftmpclass)
struct G__Definedtemplateclass *deftmpclass;
{
  if(deftmpclass->next) {
    G__freedeftemplateclass(deftmpclass->next);
    free((void*)deftmpclass->next);
    deftmpclass->next = (struct G__Definedtemplateclass *)NULL;
  }
  G__freetemplatearg(deftmpclass->def_para);
  deftmpclass->def_para=(struct G__Templatearg *)NULL;
  if(deftmpclass->name) {
    free((void*)deftmpclass->name);
    deftmpclass->name=(char*)NULL;
  }
#ifdef G__TEMPLATEMEMFUNC
  G__freetemplatememfunc(&(deftmpclass->memfunctmplt));
#endif
#ifndef G__OLDIMPLEMENTATION691
  G__IntList_free(deftmpclass->instantiatedtagnum);
  deftmpclass->instantiatedtagnum=(struct G__IntList*)NULL;
#endif
}

#ifdef G__TEMPLATEMEMFUNC
/**************************************************************************
* G__freetemplatememfunc()
**************************************************************************/
void G__freetemplatememfunc(memfunctmplt)
struct G__Definedtemplatememfunc *memfunctmplt;
{
  if(memfunctmplt->next) { 
    G__freetemplatememfunc(memfunctmplt->next);
    free((void*)memfunctmplt->next);
    memfunctmplt->next=(struct G__Definedtemplatememfunc *)NULL;
  }
}
#endif

/**************************************************************************
* G__freetemplatearg()
**************************************************************************/
void G__freetemplatearg(def_para)
struct G__Templatearg *def_para;
{
  if(def_para) {
    if(def_para->next) G__freetemplatearg(def_para->next);
    if(def_para->string) free((void*)def_para->string);
    if(def_para->default_parameter) free((void*)def_para->default_parameter);
    free((void*)def_para);
  }
}



#ifdef G__TEMPLATEFUNC
/***********************************************************************
* G__gettemplatearg()
*
*  search matches for template argument 
***********************************************************************/
char *G__gettemplatearg(n,def_para)
int n;
struct G__Templatearg *def_para;
{
  char *result;
  int i;
  G__ASSERT(def_para);
  for(i=1;i<n;i++) {
    if(def_para->next) def_para = def_para->next;
  }
  return(def_para->string);
}

/***********************************************************************
* G__istemplatearg()
*
*  search matches for template argument 
***********************************************************************/
int G__istemplatearg(paraname,def_para)
char *paraname;
struct G__Templatearg *def_para;
{
  int result=1;
  while(def_para) {
    if(strcmp(def_para->string,paraname)==0) return(result);
    def_para = def_para->next;
    ++result;
  }
  return(0);
}


/***********************************************************************
* G__checkset_charlist()
*
* Check and set actual template argument
***********************************************************************/
int G__checkset_charlist(typename,pcall_para,narg
#ifndef G__OLDIMPLEMENTATION642
			 ,ftype
#endif
			 )
char *typename;
struct G__Charlist *pcall_para;
int narg;
#ifndef G__OLDIMPLEMENTATION642
int ftype;
#endif
{
  int i;
  for(i=1;i<narg;i++) {
    if(!pcall_para->next) {
      pcall_para->next = (struct G__Charlist*)malloc(sizeof(struct G__Charlist));
      pcall_para->next->next = (struct G__Charlist*)NULL;
      pcall_para->next->string = (char*)NULL;
    }
    pcall_para = pcall_para->next;
  }

  if(pcall_para->string) {
    if(strcmp(typename,pcall_para->string)==0) return(1);
    else                                       return(0);
  }
  pcall_para->string = (char*)malloc(strlen(typename)+1);
  strcpy(pcall_para->string,typename);

#ifndef G__OLDIMPLEMENTATION642
  if('U'==ftype) {
    int len=strlen(typename);
    if(len && '*'==typename[len-1]) {
      pcall_para->string[len-1] = '\0';
    }
  }
#endif

  return(1);
}

/***********************************************************************
* G__matchtemplatefunc()
*
* Test if given function arguments and template function arguments
* matches.
***********************************************************************/
int G__matchtemplatefunc(deftmpfunc,libp,pcall_para)
struct G__Definetemplatefunc *deftmpfunc;
struct G__param *libp;
struct G__Charlist *pcall_para;
{
  int fparan,paran;
  int ftype,type;
  int ftagnum,tagnum;
  int ftypenum,typenum;
  int freftype,reftype,ref;
  int fparadefault;
  int fargtmplt;
  int i;
  char paratype[G__ONELINE];
  char *paratype2,*p;

  fparan = deftmpfunc->func_para.paran;
  paran = libp->paran;

  /* more argument in calling function, unmatch */
  if(paran>fparan) return(0);
  if(fparan>paran) {
    if(!deftmpfunc->func_para.paradefault[paran]) return(0);
  }

  for(i=0;i<paran;i++) {
    /* get template information for simplicity */
    ftype = deftmpfunc->func_para.type[i];
    ftagnum = deftmpfunc->func_para.tagnum[i];
    ftypenum = deftmpfunc->func_para.typenum[i];
    freftype = deftmpfunc->func_para.reftype[i];
    fargtmplt = deftmpfunc->func_para.argtmplt[i];

    /* get parameter information for simplicity */
    type = libp->para[i].type;
    tagnum = libp->para[i].tagnum;
    typenum = libp->para[i].typenum;
    ref = libp->para[i].ref;
    if(ref) reftype=G__PARAREFERENCE;

    /* match parameter */
    if(-1==fargtmplt) {
      if('u'!=type || -1==tagnum) return(0);
      /* template argument  (T<E> a) */
      strcpy(paratype,G__fulltagname(tagnum,0));
      paratype2 = strchr(paratype,'<');
      if(!paratype2) return(0); /* unmatch */
      *paratype2 = '\0';
      ++paratype2;
      p = strchr(paratype2,'>');
      if(!p) return(0); /* unmatch */
      *p='\0';
      /* Now, paratype="T" , paratype2="E" */
      if(ftypenum) { /*  T<c> a */
#ifndef G__OLDIMPLEMENTATION642
	if(G__checkset_charlist(paratype,pcall_para,ftypenum,ftype)) {
#else
	if(G__checkset_charlist(paratype,pcall_para,ftypenum)) {
#endif
	  /* match or newly set template argument */
	}
	else {
	  /* template argument is already set to different type, unmatch */
	  return(0);
	}
      }
      if(ftagnum) { /* c<T> a */
#ifndef G__OLDIMPLEMENTATION642
	if(G__checkset_charlist(paratype2,pcall_para,ftagnum,ftype)) {
#else
	if(G__checkset_charlist(paratype2,pcall_para,ftagnum)) {
#endif
	  /* match or newly set template argument */
	}
	else {
	  /* template argument is already set to different type, unmatch */
	  return(0);
	}
      }
    }
    else if(fargtmplt) {
      /* template argument  (T a) */
      strcpy(paratype,G__type2string(type,tagnum,typenum,0));
#ifndef G__OLDIMPLEMENTATION642
      if(G__checkset_charlist(paratype,pcall_para,fargtmplt,ftype)) {
#else
      if(G__checkset_charlist(paratype,pcall_para,fargtmplt)) {
#endif
	/* match or newly set template argument */
      }
      else {
	/* template argument is already set to different type, unmatch */
	return(0);
      }
    }
    else {
      /* fixed argument type */
      if(type==ftype&&ftagnum==tagnum&&(0==freftype||ref)) {
	/* match, check next */
      }
      else {
	/* unmatch */
	return(0);
      }
    }
  }

  return(1); /* All parameters match */

}

/***********************************************************************
* G__freetemplatefunc()
*
*
***********************************************************************/
void G__freetemplatefunc(deftmpfunc)
struct G__Definetemplatefunc *deftmpfunc;
{
  if(deftmpfunc->next) {
    G__freetemplatefunc(deftmpfunc->next);
    free((void*)deftmpfunc->next);
    deftmpfunc->next = (struct G__Definetemplatefunc*)NULL;
  }
  if(deftmpfunc->def_para) {
    G__freetemplatearg(deftmpfunc->def_para);
    deftmpfunc->def_para = (struct G__Templatearg*)NULL;
  }
  if(deftmpfunc->name) {
    free((void*)deftmpfunc->name);
    deftmpfunc->name=(char*)NULL;
  }
}

/***********************************************************************
* G__templatefunc()
*
* Search matching template function, search by name then parameter.
* If match found, expand template, parse as pre-run and execute it.
***********************************************************************/
int G__templatefunc(result,funcname,libp,hash)
G__value *result;
char *funcname;
struct G__param *libp;
int hash;
{
  struct G__Definetemplatefunc *deftmpfunc;
  struct G__Charlist call_para;
#ifndef G__OLDIMPLEMENTATION687
  int env_tagnum=G__get_envtagnum();
  struct G__inheritance *baseclass;

  if(-1!=env_tagnum) baseclass = G__struct.baseclass[env_tagnum];
  else               baseclass = &G__globalusingnamespace;
  if(0==baseclass->basen) baseclass = (struct G__inheritance*)NULL;
#endif

  call_para.string = (char*)NULL;
  call_para.next = (struct G__Charlist*)NULL;
  deftmpfunc = &G__definedtemplatefunc;
  
  /* Search matching template function name */
  while(deftmpfunc->next) {
    G__freecharlist(&call_para);
    if(deftmpfunc->hash==hash && strcmp(deftmpfunc->name,funcname)==0 &&
       G__matchtemplatefunc(deftmpfunc,libp,&call_para)) {

#ifndef G__OLDIMPLEMENTATION687
      if(-1!=deftmpfunc->parent_tagnum && 
	 env_tagnum!=deftmpfunc->parent_tagnum) {
	if(baseclass) {
	  int temp;
	  for(temp=0;temp<baseclass->basen;temp++) {
	    if(baseclass->basetagnum[temp]==deftmpfunc->parent_tagnum) {
	      goto match_found;
	    }
	  }
	}
	deftmpfunc = deftmpfunc->next;
	continue;
      }
    match_found:
#endif
      
      /* matches funcname and parameter,
       * then expand the template and parse as prerun */
      G__replacetemplate("",funcname
			 ,&call_para /* needs to make this up */
			 ,deftmpfunc->def_fp
			 ,deftmpfunc->line
			 ,deftmpfunc->filenum
			 ,&(deftmpfunc->def_pos)
			 ,deftmpfunc->def_para
			 ,0
			 ,SHRT_MAX /* large enough number */
			 );
      
      /* call the expanded template function */
      if(G__interpret_func(result,funcname,libp,hash,&G__ifunc
			   ,G__EXACT,G__TRYNORMAL)==0) {
	G__genericerror("Internal error: template function call failed");
      }
      G__freecharlist(&call_para);
      return(1); /* match */
    }
    deftmpfunc = deftmpfunc->next;
  }
  G__freecharlist(&call_para);
  return(0);  /* no match */
}
#endif /* G__TEMPLATEFUNC */

/***********************************************************************
* G__createtemplatefunc()
*
* Create template function entry
***********************************************************************/
int G__createtemplatefunc(funcname,targ,line_number,ppos)
char *funcname;
struct G__Templatearg *targ;
int line_number;
fpos_t *ppos;
{
  /*  template<class T,class E> type func(T a,E b,int a) {
   *                                      ^   */
#ifdef G__TEMPLATEFUNC
  struct G__Definetemplatefunc *deftmpfunc;
  fpos_t store_pos;
  int store_line;
  char paraname[G__MAXNAME];
  char temp[G__ONELINE];
  struct G__Templatearg *tmparg;
  int c,tmp;
  int unsigned_flag,reftype,pointlevel;
  int tagnum,typenum;
  int narg;

  /**************************************************************
  * get to the end of list 
  **************************************************************/
  deftmpfunc = &G__definedtemplatefunc;
  while(deftmpfunc->next) deftmpfunc = deftmpfunc->next;

  /**************************************************************
  * store linenumber , file pointer and file position 
  **************************************************************/
  deftmpfunc->line=line_number;
  deftmpfunc->def_pos = *ppos;
  deftmpfunc->def_fp=G__ifile.fp;
  deftmpfunc->filenum = G__ifile.filenum;

  /**************************************************************
  * store template argument list 
  **************************************************************/
  deftmpfunc->def_para = targ;

  /**************************************************************
  * store funcname and hash 
  **************************************************************/
  deftmpfunc->name=(char*)malloc(strlen(funcname)+1);
  strcpy(deftmpfunc->name,funcname);
  G__hash(funcname,deftmpfunc->hash,tmp);
#ifndef G__OLDIMPLEMENTATION687
  deftmpfunc->parent_tagnum = G__get_envtagnum();
#endif

  /**************************************************************
  * allocate next list entry
  **************************************************************/
  deftmpfunc->next 
  =(struct G__Definetemplatefunc*)malloc(sizeof(struct G__Definetemplatefunc));
  deftmpfunc->next->next = (struct G__Definetemplatefunc*)NULL;
  deftmpfunc->next->def_para = (struct G__Templatearg*)NULL;
  deftmpfunc->next->name = (char*)NULL;


  /**************************************************************
  * Parse template function parameter information 
  **************************************************************/
  /*  template<class T,class E> type func(T a,E b,int a) {
   *                                      ^   */
  deftmpfunc->func_para.paran = tmp = 0;
  c=0;
  /* read file and get type of parameter */
  while(')'!=c) {
    /* get type name */
    deftmpfunc->func_para.type[tmp] = 0;
    deftmpfunc->func_para.tagnum[tmp] = -1;
    deftmpfunc->func_para.typenum[tmp] = -1;
    deftmpfunc->func_para.reftype[tmp] = G__PARANORMAL;
    deftmpfunc->func_para.paradefault[tmp] = 0;
    deftmpfunc->func_para.argtmplt[tmp] = -1;

    unsigned_flag = reftype = pointlevel = 0;
    /*  template<class T,template<class U> class E> type func(T a,E<T> b) {
     *                                                        ^   ^  */
    do {
      c = G__fgetname_template(paraname,",)<*&=");
    } while(strcmp(paraname,"class")==0 || strcmp(paraname,"struct")==0 ||
	    strcmp(paraname,"const")==0 || strcmp(paraname,"volatile")==0);
    /*  template<class T,template<class U> class E> type func(T a,E<T> b) {
     *                                                         ^   ^  */
    /* if(isspace(c)) c = G__fgetname(temp,"<,()*&[="); */
    /*  template<class T,template<class U> class E> type func(T a,E<T> b) {
     *                                                          ^  ^  */
    if(strcmp(paraname,"unsigned")==0) {
      unsigned_flag = -1;
      if('*'!=c && '&'!=c) c = G__fgetname(paraname,",)*&=");
    }
    if(strcmp(paraname,"int")==0) {
      deftmpfunc->func_para.type[tmp] = 'i' + unsigned_flag;
    }
    else if(strcmp(paraname,"char")==0) {
      deftmpfunc->func_para.type[tmp] = 'c' + unsigned_flag;
    }
    else if(strcmp(paraname,"short")==0) {
      deftmpfunc->func_para.type[tmp] = 's' + unsigned_flag;
    }
    else if(strcmp(paraname,"long")==0) {
      deftmpfunc->func_para.type[tmp] = 'l' + unsigned_flag;
      if('*'!=c && '&'!=c) {
	c = G__fgetname(paraname,",)*&[=");
	if(strcmp(paraname,"double")==0) deftmpfunc->func_para.type[tmp]='d';
      }
    }
    else if(strcmp(paraname,"double")==0) {
      deftmpfunc->func_para.type[tmp] = 'd';
    }
    else if(strcmp(paraname,"float")==0) {
      deftmpfunc->func_para.type[tmp] = 'f';
    }
    else if(strcmp(paraname,"void")==0) {
      deftmpfunc->func_para.type[tmp] = 'y';
    }
    else if(strcmp(paraname,"FILE")==0) {
      deftmpfunc->func_para.type[tmp] = 'e';
    }
    else if(unsigned_flag) {
      deftmpfunc->func_para.type[tmp] = 'i' + unsigned_flag;
    }
    else if('<'==c) {
      /* f(T<E> a) or f(c<T> a) or f(T<c> a) */
      deftmpfunc->func_para.type[tmp]='u';
      deftmpfunc->func_para.argtmplt[tmp] = -1;
      narg=G__istemplatearg(paraname,deftmpfunc->def_para);
      deftmpfunc->func_para.typenum[tmp]=narg;
      c = G__fgetstream(paraname,">");
      narg=G__istemplatearg(paraname,deftmpfunc->def_para);
      deftmpfunc->func_para.tagnum[tmp]=narg;
    }
    else if((narg=G__istemplatearg(paraname,deftmpfunc->def_para))) {
      /* f(T a) */
#ifndef G__OLDIMPLEMENTATION642
      if('*'==c) deftmpfunc->func_para.type[tmp]='U';
      else       deftmpfunc->func_para.type[tmp]='u';
#else
      deftmpfunc->func_para.type[tmp]='u';
#endif
      deftmpfunc->func_para.argtmplt[tmp]=narg;
    }
    else {
      /* f(c a) */
      if(-1!=(typenum=G__defined_typename(paraname))) {
	deftmpfunc->func_para.type[tmp]=G__newtype.type[typenum];
	deftmpfunc->func_para.typenum[tmp]=typenum;
	deftmpfunc->func_para.tagnum[tmp]=G__newtype.tagnum[typenum];
      }
      else if(-1!=(tagnum=G__defined_tagname(paraname,0))) {
#ifndef G__OLDIMPLEMENTATION642
	/* Following 2 lines are questionable */
	if('*'==c) deftmpfunc->func_para.type[tmp]='U';
	else       deftmpfunc->func_para.type[tmp]='u';
#else
	deftmpfunc->func_para.type[tmp] = 'u'; /* How about enum ? */
#endif
	deftmpfunc->func_para.typenum[tmp] = -1;
	deftmpfunc->func_para.tagnum[tmp] = tagnum;
      }
      else {
	G__genericerror("Internal error: global function template arg type");
      }
    }

    /* Check pointlevel and reftype */
    while(','!=c && ')'!=c) {
      switch(c) {
      case '(': /* pointer to function */
	deftmpfunc->func_para.type[tmp] = 'Y'; 
	deftmpfunc->func_para.typenum[tmp] = -1;
	deftmpfunc->func_para.tagnum[tmp] = -1;
	c=G__fignorestream(")");
	c=G__fignorestream(",)");
	break;
      case '=':
	deftmpfunc->func_para.paradefault[tmp] = 1;
	c=G__fignorestream(",)");
	break;
      case '[':
	c=G__fignorestream("]");
	c = G__fgetname(temp,",()*&[=");
	++pointlevel;
	break;
      case '*':
	++pointlevel;
	c = G__fgetname(temp,",()*&[=");
	break;
      case '&':
	++reftype;
	c = G__fgetname(temp,",()*&[=");
	break;
      default:
	c = G__fgetname(temp,",()*&[=");
	break;
      }
    }
    /*  template<class T,template<class U> class E> type func(T a,E<T> b) {
     *                                                           ^      ^  */

    if(reftype) {
      if(pointlevel)
	deftmpfunc->func_para.type[tmp]=toupper(deftmpfunc->func_para.type[tmp]);
      deftmpfunc->func_para.reftype[tmp] = G__PARAREFERENCE;
    }
    else {
      switch(pointlevel) {
      case 0:
	deftmpfunc->func_para.reftype[tmp] = G__PARANORMAL ;
	break;
      case 1:
	deftmpfunc->func_para.type[tmp] = 
	  toupper(deftmpfunc->func_para.type[tmp]) ;
	deftmpfunc->func_para.reftype[tmp] = G__PARANORMAL ;
	break;
      case 2:
	deftmpfunc->func_para.type[tmp] = 
	  toupper(deftmpfunc->func_para.type[tmp]) ;
	deftmpfunc->func_para.reftype[tmp] = G__PARAP2P ;
	break;
      default:
	deftmpfunc->func_para.type[tmp] = 
	  toupper(deftmpfunc->func_para.type[tmp]) ;
	deftmpfunc->func_para.reftype[tmp] = G__PARAP2P2P ;
	break;
      }
    }

    ++tmp;
    deftmpfunc->func_para.paran = tmp;
  }

  G__fignorestream("{");
  G__fignorestream("}");

#else /* G__TEMPLATEFUNC */
  G__genericerror("Limitation: Global function template ignored");
  G__fignorestream("{");
  G__fignorestream("}");
#endif /* G__TEMPLATEFUNC */
}

#endif /* G__TEMPLATECLASS */


/*
 * Local Variables:
 * c-tab-always-indent:nil
 * c-indent-level:2
 * c-continued-statement-offset:2
 * c-brace-offset:-2
 * c-brace-imaginary-offset:0
 * c-argdecl-indent:0
 * c-label-offset:-2
 * compile-command:"make -k"
 * End:
 */
